import os
import yaml
from elasticsearch import Elasticsearch
import elasticsearch.helpers

from bel.Config import config

import logging

log = logging.getLogger(__name__)

cur_dir_name = os.path.dirname(os.path.realpath(__file__))
mappings_terms_fn = f"{cur_dir_name}/es_mappings_terms.yml"
terms_alias = "terms"


def get_all_index_names(es):
    """Get all index names"""

    indices = es.indices.get_alias()

    return indices


def add_index_alias(es, index_name, alias_name):
    """Add index alias to index_name"""

    es.indices.put_alias(index=index_name, name=terms_alias)


def index_exists(es, index_name: str):
    """Does index exist?

    Args:
        index_name:  index to check for existence
    """
    return es.indices.exists(index=index_name)


def delete_index(es, index_name: str):
    """Delete the terms index"""

    if not index_name:
        log.warn("No index name given to delete")
        return None

    result = es.indices.delete(index=index_name)
    return result


def create_terms_index(es, index_name: str):
    """Create terms index"""

    with open(mappings_terms_fn, "r") as f:
        mappings_terms = yaml.load(f, Loader=yaml.SafeLoader)

    try:
        es.indices.create(index=index_name, body=mappings_terms)

    except Exception as e:
        log.error(f"Could not create elasticsearch terms index: {e}")


def delete_terms_indexes(es, index_name: str = "terms_*"):
    """Delete all terms indexes"""

    try:
        es.indices.delete(index=index_name)
    except Exception as e:
        log.error(f"Could not delete all terms indices: {e}")


def get_client():
    """Get elasticsearch client

    Returns:
        es: Elasticsearch client handle
    """

    es = Elasticsearch([config["bel_api"]["servers"]["elasticsearch"]], send_get_body_as="POST")

    return es


def bulk_load_docs(es, docs):
    """Bulk load docs

    Args:
        es: elasticsearch handle
        docs: Iterator of doc objects - includes index_name
    """

    chunk_size = 200

    try:
        results = elasticsearch.helpers.bulk(es, docs, chunk_size=chunk_size)
        log.debug(f"Elasticsearch documents loaded: {results[0]}")

        # elasticsearch.helpers.parallel_bulk(es, terms, chunk_size=chunk_size, thread_count=4)
        if len(results[1]) > 0:
            log.error("Bulk load errors {}".format(results))
    except elasticsearch.ElasticsearchException as e:
        log.error("Indexing error: {}\n".format(e))



from typing import Mapping, Any, List, Iterable, Tuple
import gzip
import bel.lang.belobj
import jsonschema
import requests
from cityhash import CityHash64

import bel.edge.edges
from bel.Config import config

import logging

log = logging.getLogger(__name__)


# TODO is this code being used?  We also have bel.nanopub.validate.validate(nanopub, error_level) for validation


class Nanopub(object):
    """Nanopub object to manage Nanopub processing"""

    def __init__(self, endpoint: str = config.get("api", "")) -> None:
        """ Initialize Nanopub

        Args:
            endpoint (str): BEL.bio API endpoint uri, e.g. https://api.bel.bio/v1, default read from config
        """
        self.endpoint = endpoint

    def validate(
        self, nanopub: Mapping[str, Any]
    ) -> Tuple[bool, List[Tuple[str, str]]]:
        """Validates using the nanopub schema

        Args:
            nanopub (Mapping[str, Any]): nanopub dict

        Returns:
            Tuple[bool, List[Tuple[str, str]]]:
                bool: Is valid?  Yes = True, No = False
                List[Tuple[str, str]]: Validation issues, empty if valid, tuple is ('ERROR|WARNING', msg)
                    e.g. [('WARNING', "Context ID not found")]        """

        # Validate nanopub
        (is_valid, messages) = validate_to_schema(nanopub, self.nanopub_schema)
        if not is_valid:
            return messages

        # Extract BEL Version
        if nanopub["nanopub"]["type"]["name"].upper() == "BEL":
            bel_version = nanopub["nanopub"]["type"]["version"]
        else:
            is_valid = False
            return (
                is_valid,
                f"Not a BEL Nanopub according to nanopub.type.name: {nanopub['nanopub']['type']['name']}",
            )

        all_messages = []
        # Validate BEL Statements
        bel_obj = bel.lang.belobj.BEL(bel_version, self.endpoint)
        for edge in nanopub["nanopub"]["edges"]:
            bel_statement = f"{edge['subject']} {edge['relation']} {edge['object']}"
            parse_obj = bel_obj.parse(bel_statement)
            if not parse_obj.valid:
                all_messages.extend(
                    (
                        "ERROR",
                        f"BEL statement parse error {parse_obj.error}, {parse_obj.err_visual}",
                    )
                )

        # Validate nanopub.context
        for context in nanopub["nanopub"]["context"]:
            (is_valid, messages) = self.validate_context(context)
            all_messages.extend(messages)

        is_valid = True
        for _type, msg in all_messages:
            if _type == "ERROR":
                is_valid = False

        return (is_valid, all_messages)

    def validate_context(
        self, context: Mapping[str, Any]
    ) -> Tuple[bool, List[Tuple[str, str]]]:
        """ Validate context

        Args:
            context (Mapping[str, Any]): context dictionary of type, id and label

        Returns:
            Tuple[bool, List[Tuple[str, str]]]:
                bool: Is valid?  Yes = True, No = False
                List[Tuple[str, str]]: Validation issues, empty if valid, tuple is ('ERROR|WARNING', msg)
                    e.g. [('WARNING', "Context ID not found")]
        """

        url = f'{self.endpoint}/terms/{context["id"]}'

        res = requests.get(url)
        if res.status_code == 200:
            return (True, [])
        else:
            return (False, [("WARNING", f'Context {context["id"]} not found at {url}')])

    def bel_edges(
        self,
        nanopub: Mapping[str, Any],
        namespace_targets: Mapping[str, List[str]] = {},
        rules: List[str] = [],
        orthologize_target: str = None,
    ) -> List[Mapping[str, Any]]:
        """Create BEL Edges from BEL nanopub

        Args:
            nanopub (Mapping[str, Any]): bel nanopub
            namespace_targets (Mapping[str, List[str]]): what namespaces to canonicalize
            rules (List[str]): which computed edge rules to process, default is all,
               look at BEL Specification yaml file for computed edge signature keys,
               e.g. degradation, if any rule in list is 'skip', then skip computing edges
               just return primary_edge
            orthologize_target (str): species to convert BEL into, e.g. TAX:10090 for mouse, default option does not orthologize

        Returns:
            List[Mapping[str, Any]]: edge list with edge attributes (e.g. context)
        """

        edges = bel.edge.edges.create_edges(
            nanopub,
            self.endpoint,
            namespace_targets=namespace_targets,
            rules=rules,
            orthologize_target=orthologize_target,
        )

        return edges


def validate_to_schema(nanopub, schema) -> Tuple[bool, List[Tuple[str, str]]]:
    """Validate nanopub against jsonschema for nanopub

    Args:
        nanopub (Mapping[str, Any]): nanopub dict
        schema (Mapping[str, Any]): nanopub schema

    Returns:
        Tuple[bool, List[str]]:
            bool: Is valid?  Yes = True, No = False
            List[Tuple[str, str]]: Validation issues, empty if valid, tuple is ('Error|Warning', msg)
                e.g. [('ERROR', "'subject' is a required property")]
    """

    v = jsonschema.Draft4Validator(schema)
    messages = []
    errors = sorted(v.iter_errors(nanopub), key=lambda e: e.path)
    for error in errors:
        for suberror in sorted(error.context, key=lambda e: e.schema_path):
            print(list(suberror.schema_path), suberror.message, sep=", ")
            messages.append(("ERROR", suberror.message))

    is_valid = True
    if errors:
        is_valid = False

    return (is_valid, messages)


# Following is used in nanopub-tools codebase
def hash_nanopub(nanopub: Mapping[str, Any]) -> str:
    """Create CityHash64 from nanopub for duplicate check

    TODO - check that this hash value is consistent between C# and Python running on
    laptop and server

    Build string to hash

    Collect flat array of (all values.strip()):
        nanopub.type.name
        nanopub.type.version

        One of:
            nanopub.citation.database.name
            nanopub.citation.database.id

            OR

            nanopub.citation.database.uri

            OR

            nanopub.citation.database.reference

        Extend with sorted list of assertions (SRO as single string with space between S, R and O)

        Extend with sorted list of annotations (nanopub.annotations.type + ' ' + nanopub.annotations.id)

    Convert array to string by joining array elements separated by a space

    Create CityHash64(str) and return

    """

    hash_list = []

    # Type
    hash_list.append(nanopub["nanopub"]["type"].get("name", "").strip())
    hash_list.append(nanopub["nanopub"]["type"].get("version", "").strip())

    # Citation
    if nanopub["nanopub"]["citation"].get("database", False):
        hash_list.append(
            nanopub["nanopub"]["citation"]["database"].get("name", "").strip()
        )
        hash_list.append(
            nanopub["nanopub"]["citation"]["database"].get("id", "").strip()
        )
    elif nanopub["nanopub"]["citation"].get("uri", False):
        hash_list.append(nanopub["nanopub"]["citation"].get("uri", "").strip())
    elif nanopub["nanopub"]["citation"].get("reference", False):
        hash_list.append(nanopub["nanopub"]["citation"].get("reference", "").strip())

    # Assertions
    assertions = []
    for assertion in nanopub["nanopub"]["assertions"]:
        if assertion.get("relation") is None:
            assertion["relation"] = ""
        if assertion.get("object") is None:
            assertion["object"] = ""
        assertions.append(
            " ".join(
                (
                    assertion["subject"].strip(),
                    assertion.get("relation", "").strip(),
                    assertion.get("object", "").strip(),
                )
            ).strip()
        )
    assertions = sorted(assertions)
    hash_list.extend(assertions)

    # Annotations
    annotations = []

    for anno in nanopub["nanopub"]["annotations"]:
        annotations.append(
            " ".join((anno.get("type", "").strip(), anno.get("id", "").strip())).strip()
        )

    annotations = sorted(annotations)
    hash_list.extend(annotations)

    np_string = " ".join([l.lower() for l in hash_list])

    return "{:x}".format(CityHash64(np_string))



import timy
import json
import gzip
import copy

import bel.utils
import bel.db.elasticsearch as elasticsearch
import bel.db.arangodb as arangodb

import bel.resources.namespace
import bel.resources.ortholog

from bel.Config import config

from structlog import get_logger

log = get_logger()

# Set timy to track in logging mode (INFO level)
timy.timy_config.tracking_mode = timy.TrackingMode.LOGGING


def load_resource(resource_url: str, forceupdate: bool = False):
    """Load BEL Resource file

    Forceupdate will create a new index in Elasticsearch regardless of whether
    an index with the resource version already exists.

    Args:
        resource_url: URL from which to download the resource to load into the BEL API
        forceupdate: force full update - e.g. don't leave Elasticsearch indexes alone if their version ID matches
    """

    log.info(f"Loading resource {resource_url}")

    try:
        # Download resource
        fo = bel.utils.download_file(resource_url)

        if not fo:
            log.error(f"Could not download and open file {resource_url}")
            return "Failed to download resource_url"

        # Get metadata
        fo.seek(0)
        with gzip.open(fo, "rt") as f:
            metadata = json.loads(f.__next__())

        if "metadata" not in metadata:
            log.error(f"Missing metadata entry for {resource_url}")
            return "Cannot load resource file - missing metadata object in first line of file"

        # Load resource files
        if metadata["metadata"]["type"] == "namespace":
            bel.resources.namespace.load_terms(fo, metadata, forceupdate)

        elif metadata["metadata"]["type"] == "ortholog":
            bel.resources.ortholog.load_orthologs(fo, metadata)

    finally:
        fo.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Usage:  program.py <customer>

"""

import copy
import datetime
import itertools
import json
import os.path
import urllib

import requests
import structlog

import bel.db.arangodb as arangodb
import bel.edge.edges
import bel.nanopub.files as files
import bel.utils as utils

log = structlog.getLogger(__name__)

client = arangodb.get_client()
edgestore_db = arangodb.get_edgestore_handle(client)

edges_coll_name = arangodb.edgestore_edges_name
nodes_coll_name = arangodb.edgestore_nodes_name


def get_edges_for_nanopub(nanopub_id):
    query = f"""
        FOR edge IN edges
            FILTER edge.nanopub_id == "{nanopub_id}"
            LIMIT 1
            RETURN edge
    """
    try:
        result = [edge for edge in edgestore_db.aql.execute(query)]
        return result[0]
    except Exception:
        return None


def process_nanopub(
    nanopub_url, orthologize_targets: list = [], overwrite: bool = False, token: str = None
):

    log.info("Processing nanopub", nanopub_url=nanopub_url)

    url_comps = urllib.parse.urlparse(nanopub_url)
    nanopub_id = os.path.basename(url_comps.path)
    # domain = url_comps.netloc

    start_time = datetime.datetime.now()

    # collect nanopub
    headers = {}
    if token:
        headers = {"Authorization": f"Bearer {token}"}

    r = requests.get(nanopub_url, headers=headers)

    nanopub = r.json()
    log.debug(f"Nanopub {nanopub_url}", nanopub=nanopub)

    assertions = []
    for assertion in nanopub["nanopub"].get("assertions", []):
        assertions.append(f"{assertion['subject']} {assertion['relation']} {assertion['object']}")

    if not nanopub:
        return {
            "msg": f"Could not GET nanopub: {nanopub_url}",
            "edges_cnt": 0,
            "assertions_cnt": 0,
            "assertions": assertions,
            "success": False,
            "errors": [],
        }

    nanopub["source_url"] = nanopub_url

    end_time1 = datetime.datetime.now()
    delta_ms = f"{(end_time1 - start_time).total_seconds() * 1000:.1f}"
    log.debug("Timing - Get nanopub", delta_ms=delta_ms)

    # Is nanopub in edge newer than from queue? If so, skip
    if not overwrite:
        # collect one edge for nanopub from edgestore
        edge = get_edges_for_nanopub(nanopub_id)
        if edge:
            # check if edge nanopub is newer
            # log.info("Nanopub to Edge comparison", nanopub=nanopub, edge=edge)
            if edge["metadata"].get("gd_updateTS", None):
                if nanopub["nanopub"]["metadata"]["gd_updateTS"] <= edge["metadata"]["gd_updateTS"]:
                    return {"msg": "Nanopub older than edge nanopub", "success": True, "e": ""}

    end_time2 = datetime.datetime.now()
    delta_ms = f"{(end_time2 - end_time1).total_seconds() * 1000:.1f}"
    log.debug("Timing - Get edge to check nanopub", delta_ms=delta_ms)

    results = bel.edge.edges.nanopub_to_edges(nanopub, orthologize_targets=orthologize_targets)

    end_time3 = datetime.datetime.now()
    delta_ms = f"{(end_time3 - end_time2).total_seconds() * 1000:.1f}"

    if results["success"]:

        load_edges_into_db(nanopub_id, nanopub["source_url"], edges=results["edges"])

        end_time4 = datetime.datetime.now()
        delta_ms = f"{(end_time4 - end_time3).total_seconds() * 1000:.1f}"
        log.debug("Timing - Load edges into edgestore", delta_ms=delta_ms)

        delta_ms = f"{(end_time4 - start_time).total_seconds() * 1000:.1f}"
        log.debug("Timing - Process nanopub into edges", delta_ms=delta_ms)

        return {
            "msg": f"Loaded {len(results['edges'])} edges into edgestore",
            "edges_cnt": len(results["edges"]),
            "assertions_cnt": len(nanopub["nanopub"]["assertions"]),
            "assertions": assertions,
            "success": True,
            "errors": results["errors"],
        }

    else:
        return {
            "msg": f'Could not process nanopub into edges - error: {results["errors"]}',
            "edges_cnt": 0,
            "assertions_cnt": len(nanopub["nanopub"]["assertions"]),
            "assertions": assertions,
            "success": False,
            "errors": results["errors"],
        }


def load_edges_into_db(
    nanopub_id: str,
    nanopub_url: str,
    edges: list = [],
    edges_coll_name: str = edges_coll_name,
    nodes_coll_name: str = nodes_coll_name,
):
    """Load edges into Edgestore"""

    start_time = datetime.datetime.now()

    # Clean out edges for nanopub in edgestore
    query = f"""
        FOR edge IN {edges_coll_name}
            FILTER edge.nanopub_id == "{nanopub_id}"
            REMOVE edge IN edges
    """

    try:
        edgestore_db.aql.execute(query)
    except Exception as e:
        log.debug(f"Could not remove nanopub-related edges: {query}  msg: {e}")

    end_time1 = datetime.datetime.now()
    delta_ms = f"{(end_time1 - start_time).total_seconds() * 1000:.1f}"
    log.debug("Timing - Delete edges for nanopub", delta_ms=delta_ms)

    # Collect edges and nodes to load into arangodb
    node_list, edge_list = [], []
    for doc in edge_iterator(edges=edges):

        log.error("Edge to be inserted", doc=doc)

        if doc[0] == "nodes":
            node_list.append(doc[1])
        else:
            edge_list.append(doc[1])

    end_time2 = datetime.datetime.now()
    delta_ms = f"{(end_time2 - end_time1).total_seconds() * 1000:.1f}"
    log.debug("Timing - Collect edges and nodes", delta_ms=delta_ms)

    try:
        results = edgestore_db.collection(edges_coll_name).import_bulk(
            edge_list, on_duplicate="replace", halt_on_error=False
        )
    except Exception as e:
        log.error(f"Could not load edges  msg: {e}")

    end_time3 = datetime.datetime.now()
    delta_ms = f"{(end_time3 - end_time2).total_seconds() * 1000:.1f}"
    log.debug("Timing - Load edges into edgestore", delta_ms=delta_ms)

    try:
        results = edgestore_db.collection(nodes_coll_name).import_bulk(
            node_list, on_duplicate="replace", halt_on_error=False
        )
    except Exception as e:
        log.error(f"Could not load nodes  msg: {e}")

    end_time4 = datetime.datetime.now()
    delta_ms = f"{(end_time4 - end_time3).total_seconds() * 1000:.1f}"
    log.debug("Timing - Load nodes into edgestore", delta_ms=delta_ms)


def edge_iterator(edges=[], edges_fn=None):
    """Yield documents from edge for loading into ArangoDB"""

    for edge in itertools.chain(edges, files.read_edges(edges_fn)):

        subj = copy.deepcopy(edge["edge"]["subject"])
        subj_id = str(utils._create_hash_from_doc(subj))
        subj["_key"] = subj_id
        obj = copy.deepcopy(edge["edge"]["object"])

        obj_id = str(utils._create_hash_from_doc(obj))
        obj["_key"] = obj_id
        relation = copy.deepcopy(edge["edge"]["relation"])

        relation["_from"] = f"nodes/{subj_id}"
        relation["_to"] = f"nodes/{obj_id}"

        # Create edge _key
        relation_hash = copy.deepcopy(relation)
        relation_hash.pop("edge_dt", None)
        relation_hash.pop("edge_hash", None)
        relation_hash.pop("nanopub_dt", None)
        relation_hash.pop("nanopub_url", None)
        relation_hash.pop("subject_canon", None)
        relation_hash.pop("object_canon", None)
        relation_hash.pop("public_flag", None)
        relation_hash.pop("metadata", None)

        relation_id = str(utils._create_hash_from_doc(relation_hash))
        relation["_key"] = relation_id

        if edge.get("nanopub_id", None):
            if "metadata" not in relation:
                relation["metadata"] = {}
            relation["metadata"]["nanopub_id"] = edge["nanopub_id"]

        yield ("nodes", subj)
        yield ("nodes", obj)
        yield ("edges", relation)


def main():
    pass


if __name__ == "__main__":
    main()



import importlib
import sys
import os
import datetime
from typing import Mapping, Any, List, Union
from tatsu.exceptions import FailedParse

import bel.lang.bel_utils as bel_utils
import bel.lang.bel_specification as bel_specification
import bel.lang.ast as lang_ast
import bel.lang.exceptions as bel_ex
import bel.lang.semantics as semantics
import bel.edge.computed

from bel.Config import config

import structlog

log = structlog.getLogger(__name__)

sys.path.append("../")


class BEL(object):
    """BEL Language object

    This object handles BEL statement/triple processing, parsing, (de)canonicalization,
    orthologization, computing BEL Edges and (TODO) statement completion.

    To convert BEL Statement to BEL Edges:

        statement = "p(HGNC:AKT1) increases p(HGNC:EGF)"
        bel_obj = bel.lang.belobj.BEL('2.0.0', 'https://api.bel.bio/v1')  # can get default version and api_url from belbio_conf.yml file as well
        bel_obj.parse(statement)  # Adds ast to bel_obj
        bel_obj.orthologize('TAX:10090')  # Run orthologize before canonicalize if needed, updates bel_obj.ast and returns self
        bel_obj.canonicalize()  # updates bel_obj.ast and returns self

        computed_edges = bel_obj.computed()

        primary_edge = bel_obj.ast.to_triple()
    """

    def __init__(self, version: str = None, api_url: str = None) -> None:
        """Initialize BEL object used for validating/processing/etc BEL statements

        Args:
            version (str): BEL Version, defaults to config['bel']['lang']['default_bel_version']
            api_url (str): BEL API endpoint,  defaults to config['bel_api']['servers']['api_url']
        """

        bel_versions = bel_specification.get_bel_versions()

        # use bel_utils._default_to_version to check if valid version, and if it exists or not
        if not version:
            self.version = config["bel"]["lang"]["default_bel_version"]
        else:
            self.version = version

        if self.version not in bel_versions:
            log.warning(
                f"Cannot validate with invalid version: {self.version} in BEL Versions: {bel_versions}"
            )

        if not api_url:
            self.api_url = config["bel_api"]["servers"]["api_url"]
        else:
            self.api_url = api_url

        # Validation error/warning messages
        # List[Tuple[str, str]], e.g. [('ERROR', 'this is an error msg'), ('WARNING', 'this is a warning'), ]
        self.validation_messages = []

        # self.semantics = BELSemantics()  # each instance also instantiates a BELSemantics object used in parsing statements
        self.spec = bel_specification.get_specification(self.version)

        # bel_utils._dump_spec(self.spec)

        # Import Tatsu parser
        # use importlib to import our parser (a .py file) and set the BELParse object as an instance variable
        try:
            parser_fn = self.spec["admin"]["parser_fn"]

            parser_name = os.path.basename(parser_fn).replace(".py", "")
            module_spec = importlib.util.spec_from_file_location(parser_name, parser_fn)
            imported_parser = importlib.util.module_from_spec(module_spec)
            module_spec.loader.exec_module(imported_parser)
            self.parser = imported_parser.BELParser()
        except Exception as e:
            # if not found, we raise the NoParserFound exception which can be found in bel.lang.exceptions
            raise bel_ex.NoParserFound(f"Version: {self.version} Msg: {e}")

    def parse(
        self,
        assertion: Union[str, Mapping[str, str]],
        strict: bool = False,
        parseinfo: bool = False,
        rule_name: str = "start",
        error_level: str = "WARNING",
    ) -> "BEL":
        """Parse and semantically validate BEL statement

        Parses a BEL statement given as a string and returns an AST, Abstract Syntax Tree (defined in ast.py)
        if the statement is valid, self.parse_valid. Else, the AST attribute is None and there will be validation error messages
        in self.validation_messages.  self.validation_messages will contain WARNINGS if
        warranted even if the statement parses correctly.

        Error Levels are similar to log levels - selecting WARNING includes both
        WARNING and ERROR, selecting ERROR just includes ERROR

        Args:
            assertion: BEL statement (if str -> 'S R O', if dict {'subject': S, 'relation': R, 'object': O})
            strict: specify to use strict or loose parsing; defaults to loose
            parseinfo: specify whether or not to include Tatsu parse information in AST
            rule_name: starting point in parser - defaults to 'start'
            error_level: return ERRORs only or also WARNINGs

        Returns:
            ParseObject: The ParseObject which contain either an AST or error messages.
        """

        self.ast = None
        self.parse_valid = False
        self.parse_visualize_error = ""
        self.validation_messages = []  # Reset messages when parsing a new BEL Statement

        if isinstance(assertion, dict):
            if assertion.get("relation", False) and assertion.get("object", False):
                statement = f"{assertion['subject']} {assertion['relation']} {assertion['object']}"
            elif assertion.get("subject"):
                statement = f"{assertion['subject']}"
            else:
                statement = ""
        else:
            statement = assertion

        self.original_bel_stmt = statement

        # pre-process to remove extra white space, add space after commas, etc.
        self.bel_stmt = bel_utils.preprocess_bel_stmt(statement)

        # TODO - double check these tests before enabling
        # is_valid, messages = bel_utils.simple_checks(self.bel_stmt)
        # if not is_valid:
        #     self.validation_messages.extend(messages)
        #     return self

        # Check to see if empty string for bel statement
        if len(self.bel_stmt) == 0:
            self.validation_messages.append(
                ("ERROR", "Please include a valid BEL statement - found empty string.")
            )
            return self

        try:
            # see if an AST is returned without any parsing errors

            ast_dict = self.parser.parse(
                self.bel_stmt, rule_name=rule_name, trace=False, parseinfo=parseinfo
            )
            self.ast = lang_ast.ast_dict_to_objects(ast_dict, self)

            self.parse_valid = True

        except FailedParse as e:
            # if an error is returned, send to handle_syntax, error
            error, visualize_error = bel_utils.handle_parser_syntax_error(e)
            self.parse_visualize_error = visualize_error
            if visualize_error:
                self.validation_messages.append(("ERROR", f"{error}\n{visualize_error}"))
            else:
                self.validation_messages.append(("ERROR", f"{error}\nBEL: {self.bel_stmt}"))
            self.ast = None

        except Exception as e:
            log.error("Error {}, error type: {}".format(e, type(e)))
            self.validation_messages.append(
                ("ERROR", "Error {}, error type: {}".format(e, type(e)))
            )

        return self

    def semantic_validation(self, error_level: str = "WARNING") -> "BEL":
        """Semantically validate parsed BEL statement

        Run semantics validation - and decorate AST with nsarg entity_type and arg optionality

        Args:
            error_level:  WARNING or ERROR

        Returns:
            BEL: return self
        """

        semantics.validate(self, error_level)

        return self

    def canonicalize(self, namespace_targets: Mapping[str, List[str]] = None) -> "BEL":
        """
        Takes an AST and returns a canonicalized BEL statement string.

        Args:
            namespace_targets (Mapping[str, List[str]]): override default canonicalization
                settings of BEL.bio API api_url - see {api_url}/status to get default canonicalization settings

        Returns:
            BEL: returns self
        """

        # TODO Need to order position independent args

        if not self.ast:
            return self

        # Collect canonical/decanonical NSArg values
        if not self.ast.collected_nsarg_norms:
            self = self.collect_nsarg_norms()

        # TODO Need to pass namespace target overrides for canonicalization
        self.ast.canonicalize()

        # self.ast = bel_utils.convert_namespaces_ast(self.ast, canonicalize=True, api_url=self.api_url, namespace_targets=namespace_targets)

        return self

    def decanonicalize(self, namespace_targets: Mapping[str, List[str]] = None) -> "BEL":
        """
        Takes an AST and returns a decanonicalized BEL statement string.

        Args:
            namespace_targets (Mapping[str, List[str]]): override default decanonicalization
                settings of BEL.bio API endpoint - see {endpoint}/status to get default decanonicalization settings

        Returns:
            BEL: returns self
        """

        if not self.ast:
            return self

        # Collect canonical/decanonical NSArg values
        if not self.ast.collected_nsarg_norms:
            self = self.collect_nsarg_norms()

        self.ast.decanonicalize()

        # self.ast = bel_utils.convert_namespaces_ast(self.ast, decanonicalize=True, namespace_targets=namespace_targets)
        return self

    def collect_nsarg_norms(self):
        """Adds canonical and decanonical values to NSArgs in AST

        This prepares the AST object for (de)canonicalization
        """
        start_time = datetime.datetime.now()

        self.ast = bel_utils.populate_ast_nsarg_defaults(self.ast, self.ast)
        self.ast.collected_nsarg_norms = True
        if (
            hasattr(self.ast, "bel_object")
            and self.ast.bel_object
            and self.ast.bel_object.type == "BELAst"
        ):
            self.ast.bel_object.collected_nsarg_norms = True

        end_time = datetime.datetime.now()
        delta_ms = f"{(end_time - start_time).total_seconds() * 1000:.1f}"

        log.debug("Timing - prepare nsarg normalization", delta_ms=delta_ms)

        return self

    def orthologize(self, species_id: str) -> "BEL":
        """Orthologize BEL AST to given species_id

        Will return original entity (ns:value) if no ortholog found.

        Args:
            species_id (str): species id to convert genes/rna/proteins into

        Returns:
            BEL: returns self
        """

        if not self.ast:
            return self

        # Collect canonical/decanonical NSArg values
        if not self.ast.collected_orthologs:
            self = self.collect_orthologs([species_id])

        self.ast.species = set()
        self.ast = bel_utils.orthologize(self.ast, self, species_id)

        return self

    def collect_orthologs(self, species: list) -> "BEL":
        """Add NSArg orthologs for given species (TAX:<number format)

        This will add orthologs to the AST for all of the given species (as available).

        NOTE: This will run self.collect_nsarg_norms() first if not already available
        as we need the canonical forms of the NSArgs
        """

        if not species:
            return self

        species_labels = bel.terms.terms.get_labels(species)

        # Collect canonical/decanonical NSArg values
        if not self.ast.collected_nsarg_norms:
            self = self.collect_nsarg_norms()

        start_time = datetime.datetime.now()

        self.ast = bel_utils.populate_ast_nsarg_orthologs(self.ast, species_labels)
        self.ast.collected_orthologs = True
        if (
            hasattr(self.ast, "bel_object")
            and self.ast.bel_object
            and self.ast.bel_object.type == "BELAst"
        ):
            self.ast.bel_object.collected_orthologs = True

        end_time = datetime.datetime.now()
        delta_ms = f"{(end_time - start_time).total_seconds() * 1000:.1f}"

        log.debug("Timing - prepare nsarg normalization", delta_ms=delta_ms)

        return self

    def compute_edges(
        self, rules: List[str] = None, ast_result=False, fmt="medium"
    ) -> List[Mapping[str, Any]]:
        """Computed edges from primary BEL statement

        Takes an AST and generates all computed edges based on BEL Specification YAML computed signatures.
        Will run only the list of computed edge rules if given.

        Args:
            rules (list): a list of rules to filter; only the rules in this list will be applied to computed
            fmt (str): short, medium or long version of BEL Edge (function and relation names)
        Returns:
            List[Mapping[str, Any]]: BEL Edges in medium format
        """

        if not self.ast:
            return self

        edges_asts = bel.edge.computed.compute_edges(self.ast, self.spec)

        if ast_result:
            return edges_asts

        edges = []
        for ast in edges_asts:
            edges.append(
                {
                    "subject": ast.bel_subject.to_string(),
                    "relation": ast.bel_relation,
                    "object": ast.bel_object.to_string(),
                }
            )

        return edges

    def to_string(self, fmt: str = "medium") -> str:
        """Convert AST object to string

        Args:
            fmt (str): short, medium, long formatted BEL statements
                short = short function and short relation format
                medium = short function and long relation format
                long = long function and long relation format

        Returns:
            str: string version of BEL AST
        """

        if self.ast:
            return f"{self.ast.to_string(ast_obj=self.ast, fmt=fmt)}"

    def to_triple(self, fmt: str = "medium") -> dict:
        """Convert AST object to BEL triple

        Args:
            fmt (str): short, medium, long formatted BEL statements
                short = short function and short relation format
                medium = short function and long relation format
                long = long function and long relation format

        Returns:
            dict: {'subject': <subject>, 'relation': <relations>, 'object': <object>}
        """

        if self.ast:
            return self.ast.to_triple(ast_obj=self.ast, fmt=fmt)
        else:
            return {}

    def print_tree(self) -> str:
        """Convert AST object to tree view of BEL AST

        Returns:
            printed tree of BEL AST
        """

        if self.ast:
            return self.ast.print_tree(ast_obj=self.ast)
        else:
            return ""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Usage:  import bel.Config as config

Get first belbio_conf.{yml|yaml} and belbio_secrets.{yml|yaml} files in:
    current dir or parent directories starting with immediate parent of current dir
    OR
    ~/.belbio_{conf|secrets}   (dotfiles in home directory)
"""

import os
import re
import yaml
import copy
from typing import Mapping, Any, MutableMapping
import functools
import collections

import logging

log = logging.getLogger(__name__)


def get_belbio_conf_files():
    """Get belbio configuration from files
    """

    home = os.path.expanduser("~")
    cwd = os.getcwd()

    belbio_conf_fp, belbio_secrets_fp = "", ""

    env_conf_dir = os.getenv("BELBIO_CONF", "").rstrip("/")

    conf_paths = [
        f"{cwd}/belbio_conf.yaml",
        f"{cwd}/belbio_conf.yml",
        f"{env_conf_dir}/belbio_conf.yaml",
        f"{env_conf_dir}/belbio_conf.yml",
        f"{home}/.belbio/conf",
    ]
    secret_paths = [
        f"{cwd}/belbio_secrets.yaml",
        f"{cwd}/belbio_secrets.yml",
        f"{env_conf_dir}/belbio_secrets.yaml",
        f"{env_conf_dir}/belbio_secrets.yml",
        f"{home}/.belbio/secrets",
    ]

    for fn in conf_paths:
        if os.path.exists(fn):
            belbio_conf_fp = fn
            break
    else:
        log.error(
            "No BELBio configuration file found - please add one (see http://bel.readthedocs.io/en/latest/configuration.html)"
        )

    for fn in secret_paths:
        if os.path.exists(fn):
            belbio_secrets_fp = fn
            break

    return (belbio_conf_fp, belbio_secrets_fp)


def load_configuration():
    """Load the configuration"""

    (belbio_conf_fp, belbio_secrets_fp) = get_belbio_conf_files()
    log.info(f"Using conf: {belbio_conf_fp} and secrets files: {belbio_secrets_fp} ")

    config = {}
    if belbio_conf_fp:
        with open(belbio_conf_fp, "r") as f:
            config = yaml.load(f, Loader=yaml.SafeLoader)
            config["source_files"] = {}
            config["source_files"]["conf"] = belbio_conf_fp

    if belbio_secrets_fp:
        with open(belbio_secrets_fp, "r") as f:
            secrets = yaml.load(f, Loader=yaml.SafeLoader)
            config["secrets"] = copy.deepcopy(secrets)
            if "source_files" in config:
                config["source_files"]["secrets"] = belbio_secrets_fp

    get_versions(config)

    # TODO - needs to be completed
    # add_environment_vars(config)

    return config


def get_versions(config) -> dict:
    """Get versions of bel modules and tools"""

    # Collect bel package version
    try:
        import bel.__version__

        config["bel"]["version"] = bel.__version__.__version__
    except KeyError:
        config["bel"] = {"version": bel.__version__.__version__}
    except ModuleNotFoundError:
        pass

    # Collect bel_resources version
    try:
        import tools.__version__

        config["bel_resources"]["version"] = tools.__version__.__version__
    except KeyError:
        config["bel_resources"] = {"version": tools.__version__.__version__}
    except ModuleNotFoundError:
        pass

    # Collect bel_api version
    try:
        import __version__

        if __version__.__name__ == "BELBIO API":
            config["bel_api"]["version"] = __version__.__version__
    except KeyError:
        if __version__.__name__ == "BELBIO API":
            config["bel_api"] = {"version": __version__.__version__}
    except ModuleNotFoundError:
        pass


# TODO - still needs to be completed
def add_environment_vars(config: MutableMapping[str, Any]):
    """Override config with environment variables

    Environment variables have to be prefixed with BELBIO_
    which will be stripped before splitting on '__' and lower-casing
    the environment variable name that is left into keys for the
    config dictionary.

    Example:
        BELBIO_BEL_API__SERVERS__API_URL=http://api.bel.bio
        1. BELBIO_BEL_API__SERVERS__API_URL ==> BEL_API__SERVERS__API_URL
        2. BEL_API__SERVERS__API_URL ==> bel_api__servers__api_url
        3. bel_api__servers__api_url ==> [bel_api, servers, api_url]
        4. [bel_api, servers, api_url] ==> config['bel_api']['servers']['api_url'] = http://api.bel.bio

    """

    # TODO need to redo config - can't add value to dictionary without recursively building up the dict
    #         check into config libraries again

    for e in os.environ:
        if re.match("BELBIO_", e):
            val = os.environ.get(e)
            if val:
                e.replace("BELBIO_", "")
                env_keys = e.lower().split("__")
                if len(env_keys) > 1:
                    joined = '"]["'.join(env_keys)
                    eval_config = f'config["{joined}"] = val'
                    try:
                        eval(eval_config)
                    except Exception as exc:
                        log.warn("Cannot process {e} into config")
                else:
                    config[env_keys[0]] = val


def merge_config(
    config: Mapping[str, Any],
    override_config: Mapping[str, Any] = None,
    override_config_fn: str = None,
) -> Mapping[str, Any]:
    """Override config with additional configuration in override_config or override_config_fn

    Used in script to merge CLI options with Config

    Args:
        config: original configuration
        override_config: new configuration to override/extend current config
        override_config_fn: new configuration filename as YAML file
    """

    if override_config_fn:
        with open(override_config_fn, "r") as f:
            override_config = yaml.load(f, Loader=yaml.SafeLoader)

    if not override_config:
        log.info("Missing override_config")

    return functools.reduce(rec_merge, (config, override_config))


# https://stackoverflow.com/questions/7204805/dictionaries-of-dictionaries-merge
def rec_merge(d1, d2):
    """ Recursively merge two dictionaries

    Update two dicts of dicts recursively,
    if either mapping has leaves that are non-dicts,
    the second's leaf overwrites the first's.

    import collections
    import functools

    e.g. functools.reduce(rec_merge, (d1, d2, d3, d4))
    """

    for k, v in d1.items():
        if k in d2:
            # this next check is the only difference!
            if all(isinstance(e, collections.MutableMapping) for e in (v, d2[k])):
                d2[k] = rec_merge(v, d2[k])
            # we could further check types and merge as appropriate here.
    d3 = d1.copy()
    d3.update(d2)
    return d3


def main():

    import json

    config = load_configuration()
    print("Config:\n", json.dumps(load_configuration(), indent=4))
    print("ns def", config["bel_resources"]["file_locations"]["namespaces_definition"])


if __name__ == "__main__":
    main()

else:
    # If building documents in readthedocs - provide empty config
    if os.getenv("READTHEDOCS", False):
        config = {}
        log.info("READTHEDOCS environment")
    else:
        config = load_configuration()


# Ideas for providing hierarchical settings
# https://stackoverflow.com/questions/7204805/dictionaries-of-dictionaries-merge
# def dict_merge(a, b):
#     if not isinstance(b, dict):
#         return b
#     result = copy.deepcopy(a)
#     for k, v in b.iteritems():
#         if k in result and isinstance(result[k], dict):
#                 result[k] = dict_merge(result[k], v)
#         else:
#             result[k] = copy.deepcopy(v)
#     return result



########################
# BEL FUNCTION IMPORTS #
########################

import bel.db
import bel.edge
import bel.lang
import bel.nanopub
import bel.resources

from bel.lang.belobj import BEL
import bel.lang.bel_specification as bel_specification

import logging

logging.getLogger(__name__).addHandler(logging.NullHandler())



class InvalidArgeter(Exception):
    def __init__(self, message):
        self.message = message


class ArgeterMissing(Exception):
    def __init__(self, function):
        self.message = "There are no args given to {}().".format(function)


class NoValidSignature(Exception):
    def __init__(self, message):
        self.message = message


class InvalidRelationship(Exception):
    def __init__(self, given_r):
        self.message = '"{}" is not a defined relation.'.format(given_r)


class MissingParenthesis(Exception):
    def __init__(self, message):
        self.message = message


class InvalidCharacter(Exception):
    def __init__(self, message):
        self.message = message


class MissingQuotation(Exception):
    def __init__(self, message):
        self.message = message


class NoParserFound(Exception):
    def __init__(self, version):
        self.message = "No parser found for BEL v{}!".format(version)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Completion notes
1. Won't complete inside of a quote.  If there are mis-matched quotes
   it will break completions
2. terms upstream of a '(' are functions/modifiers
3. commas separate arguments of upstream function
"""

from typing import Mapping, Any, List, Tuple, Optional
import re
import json
import copy

from bel.utils import get_url, url_path_param_quoting

import bel.lang.partialparse as pparse
import bel.lang.bel_specification as bel_specification

from bel.Config import config

from structlog import get_logger

log = get_logger()

default_bel_version = config["bel"]["lang"]["default_bel_version"]

# Custom Typing definitions
Span = Tuple[int, int]
AST = Mapping[str, Any]
BELSpec = Mapping[str, Any]


def in_span(loc: int, span: Span) -> bool:
    """Checks if loc is inside span"""

    if loc >= span[0] and loc <= span[1]:
        return True
    else:
        return False


def cursor(
    belstr: str, ast: AST, cursor_loc: int, result: Mapping[str, Any] = None
) -> Mapping[str, Any]:
    """Find BEL function or argument at cursor location

    Args:
        belstr: BEL String used to create the completion_text
        ast (Mapping[str, Any]): AST (dict) of BEL String
        cursor_loc (int): given cursor location from input field
            cursor_loc starts at 0, think of it like a block cursor covering each char
        result: used to recursively return the result
    Returns:
        result dict
    """

    log.debug(f"SubAST: {json.dumps(ast, indent=4)}")

    # Recurse down through subject, object, nested to functions
    log.debug(f"Cursor keys {ast.keys()}, BELStr: {belstr}")

    if len(belstr) == 0:
        return {"type": "Function", "replace_span": (0, 0), "completion_text": ""}

    if "relation" in ast and in_span(cursor_loc, ast["relation"]["span"]):
        log.debug("In relation")

        completion_text = belstr[ast["relation"]["span"][0] : cursor_loc + 1]
        return {
            "type": "Relation",
            "replace_span": ast["relation"]["span"],
            "completion_text": completion_text,
        }

    # Handle subject, object and nested keys in tree
    elif "span" not in ast and isinstance(ast, dict):
        for key in ast:
            if key in ["subject", "object", "nested"]:
                log.debug(f"Recursing Keys {key}")
                result = cursor(belstr, ast[key], cursor_loc, result=result)
                if result:
                    return result

    # Matches Functions, NSArgs and StrArgs/StrArgNSArg
    if "span" in ast and in_span(cursor_loc, ast["span"]):
        log.debug("Inside subject/object subAST")
        if "function" in ast:
            name_span = ast["function"]["name_span"]
            if in_span(cursor_loc, name_span):
                return {
                    "type": "Function",
                    "replace_span": name_span,
                    "completion_text": belstr[name_span[0] : cursor_loc + 1],
                }
            for idx, arg in enumerate(ast["args"]):
                if (
                    cursor_loc == ast["function"]["parens_span"][0]
                    and ast["function"]["parens_span"][1] == -1
                ):
                    return {
                        "type": "StrArg",  # default type if unknown
                        "arg_idx": idx,
                        "replace_span": arg["span"],  # replace entire strarg
                        "parent_function": ast["function"]["name"],
                        "completion_text": "",
                    }

                elif in_span(cursor_loc, arg["span"]):
                    log.debug(
                        f'In argument span {arg["span"]}  Cursor_loc: {cursor_loc}'
                    )
                    if arg["type"] == "Function":
                        if in_span(cursor_loc, arg["function"]["name_span"]):
                            log.debug("Found replace_span in args: Function type")
                            return {
                                "type": "Function",
                                "replace_span": arg["function"][
                                    "name_span"
                                ],  # replace function name only
                                "arg_idx": idx,
                                "args": copy.deepcopy(ast["args"]),
                                "parent_function": ast["function"]["name"],
                                "completion_text": belstr[
                                    arg["function"]["name_span"][0] : cursor_loc + 1
                                ],
                            }
                        else:
                            log.debug(f'Recursing Function  {arg["span"]}')
                            result = cursor(belstr, arg, cursor_loc, result=result)
                    elif arg["type"] == "NSArg":

                        result = {
                            "type": "NSArg",
                            "replace_span": arg["span"],  # replace entire nsarg
                            "arg_idx": idx,
                            "args": copy.deepcopy(ast["args"]),
                            "parent_function": ast["function"]["name"],
                        }

                        # Filter on namespace and query on ns_val chars up to cursor_loc
                        if in_span(cursor_loc, arg["nsarg"]["ns_val_span"]):
                            result["namespace"] = arg["nsarg"][
                                "ns"
                            ]  # provide namespace for nsarg filtering
                            result["completion_text"] = belstr[
                                arg["nsarg"]["ns_val_span"][0] : cursor_loc + 1
                            ]
                        # Query on nsarg chars up to cursor_loc
                        else:
                            result["completion_text"] = belstr[
                                arg["nsarg"]["ns_span"][0] : cursor_loc + 1
                            ]

                        log.debug(f"Found replace_span in args: NSArg {result}")
                        return result
                    elif (
                        arg["type"] == "StrArg"
                    ):  # in case this is a default namespace StrArg
                        if (
                            arg["span"][0] == arg["span"][1]
                        ):  # handle case like p() cursor=2
                            completion_text = arg["arg"]
                        else:
                            completion_text = belstr[arg["span"][0] : cursor_loc + 1]

                        return {
                            "type": "StrArg",
                            "arg_idx": idx,
                            "replace_span": arg["span"],  # replace entire strarg
                            "parent_function": ast["function"]["name"],
                            "completion_text": completion_text.lstrip(),
                        }
    return result  # needed to pass result back up recursive stack


def nsarg_completions(
    completion_text: str,
    entity_types: list,
    bel_spec: BELSpec,
    namespace: str,
    species_id: str,
    bel_fmt: str,
    size: int,
):
    """Namespace completions

    Args:
        completion_text
        entity_types: used to filter namespace search results
        bel_spec: used to search default namespaces
        namespace: used to filter namespace search results
        species_id: used to filter namespace search results
        bel_fmt: used to select full name or abbrev for default namespaces
        size: how many completions to return

    Results:
        list of replacement text objects
    """

    minimal_nsarg_completion_len = 1

    species = [species_id]
    namespaces = [namespace]
    replace_list = []

    if len(completion_text) >= minimal_nsarg_completion_len:
        # Use BEL.bio API module if running bel module in BEL.bio API, otherwise call BEL.bio API endpoint
        #   is there a better way to  handle this?

        url = f'{config["bel_api"]["servers"]["api_url"]}/terms/completions/{url_path_param_quoting(completion_text)}'
        params = {
            "size": size,
            "entity_types": entity_types,
            "namespaces": namespaces,
            "species": species,
        }
        r = get_url(url, params=params)

        if r.status_code == 200:
            ns_completions = r.json()
        else:
            log.error(f"Status code of {r.status_code} for {url}")
            ns_completions = {}

        for complete in ns_completions.get("completions", []):
            replace_list.append(
                {
                    "replacement": complete["id"],
                    "label": f"{complete['id']} ({complete['label']})",
                    "highlight": complete["highlight"][-1],
                    "type": "NSArg",
                }
            )

    # Check default namespaces
    for entity_type in entity_types:
        default_namespace = bel_spec["namespaces"].get(entity_type, [])
        if default_namespace:
            for obj in default_namespace["info"]:
                replacement = None
                if bel_fmt == "long" and re.match(
                    completion_text, obj["name"], re.IGNORECASE
                ):
                    replacement = obj["name"]
                elif bel_fmt in ["short", "medium"] and re.match(
                    completion_text, obj["abbreviation"], re.IGNORECASE
                ):
                    replacement = obj["abbreviation"]

                if replacement:
                    highlight = replacement.replace(
                        completion_text, f"<em>{completion_text}</em>"
                    )
                    replace_list.insert(
                        0,
                        {
                            "replacement": replacement,
                            "label": replacement,
                            "highlight": highlight,
                            "type": "NSArg",
                        },
                    )

    return replace_list[:size]


def relation_completions(
    completion_text: str, bel_spec: BELSpec, bel_fmt: str, size: int
) -> list:
    """Filter BEL relations by prefix

    Args:
        prefix: completion string
        bel_fmt: short, medium, long BEL formats
        spec: BEL specification

    Returns:
        list: list of BEL relations that match prefix
    """

    if bel_fmt == "short":
        relation_list = bel_spec["relations"]["list_short"]
    else:
        relation_list = bel_spec["relations"]["list_long"]

    matches = []
    for r in relation_list:
        if re.match(completion_text, r):
            matches.append(r)

    replace_list = []
    for match in matches:
        highlight = match.replace(completion_text, f"<em>{completion_text}</em>")
        replace_list.append(
            {
                "replacement": match,
                "label": match,
                "highlight": highlight,
                "type": "Relation",
            }
        )

    return replace_list[:size]


def function_completions(
    completion_text: str,
    bel_spec: BELSpec,
    function_list: list,
    bel_fmt: str,
    size: int,
) -> list:
    """Filter BEL functions by prefix

    Args:
        prefix: completion string
        bel_fmt: short, medium, long BEL formats
        spec: BEL specification

    Returns:
        list: list of BEL functions that match prefix
    """

    # Convert provided function list to correct bel_fmt
    if isinstance(function_list, list):
        if bel_fmt in ["short", "medium"]:
            function_list = [
                bel_spec["functions"]["to_short"][fn] for fn in function_list
            ]
        else:
            function_list = [
                bel_spec["functions"]["to_long"][fn] for fn in function_list
            ]
    elif bel_fmt in ["short", "medium"]:
        function_list = bel_spec["functions"]["primary"]["list_short"]
    else:
        function_list = bel_spec["functions"]["primary"]["list_long"]

    matches = []
    for f in function_list:
        escaped_completion_text = completion_text.replace(r"(", r"\(").replace(
            r")", r"\)"
        )
        log.debug(f"Completion match: {escaped_completion_text}  F: {f}")
        if re.match(escaped_completion_text, f):
            matches.append(f)

    replace_list = []
    for match in matches:
        if completion_text:
            highlight = match.replace(completion_text, f"<em>{completion_text}</em>")
        else:
            highlight = completion_text

        replace_list.append(
            {
                "replacement": match,
                "label": f"{match}()",
                "highlight": highlight,
                "type": "Function",
            }
        )

    return replace_list[:size]


def arg_completions(
    completion_text: str,
    parent_function: str,
    args: list,
    arg_idx: int,
    bel_spec: BELSpec,
    bel_fmt: str,
    species_id: str,
    namespace: str,
    size: int,
):
    """Function argument completion

    Only allow legal options for completion given function name, arguments and index of argument
    to replace.

    Args:
        completion_text: text to use for completion - used for creating highlight
        parent_function: BEL function containing these args
        args: arguments of BEL function
        arg_idx: completing on this argument identified by this index
        bel_spec: BEL Specification
        bel_fmt: short, medium, long BEL function/relation formats
        species_id: filter on this species id, e.g. TAX:9606 if available
        namespace: filter on this namespace if available
        size: number of completions to return

    Return:
        list of replacements
    """

    function_long = bel_spec["functions"]["to_long"].get(parent_function)
    if not function_long:
        return []

    signatures = bel_spec["functions"]["signatures"][function_long]["signatures"]

    # Position based argument  ###################################
    function_list = []
    entity_types = []
    fn_replace_list, ns_arg_replace_list = [], []
    position_flag = False  # Signature matches position-based argument

    # Check for position based argument
    for signature in signatures:
        sig_arg = signature["arguments"][arg_idx]
        sig_type = sig_arg["type"]

        if sig_arg.get("position", False) and arg_idx == sig_arg["position"] - 1:
            position_flag = True
            if sig_type in ["Function", "Modifier"]:
                function_list.extend(sig_arg["values"])
            elif sig_type in ["NSArg", "StrArgNSArg"]:
                entity_types.extend(sig_arg["values"])

    if not position_flag:
        # Collect optional and multiple signature arguments for completion
        opt_fn_sig_args = []
        opt_nsarg_sig_args = []
        mult_fn_sig_args = []
        mult_nsarg_sig_args = []
        for signature in signatures:
            signature_opt_fn_sig_args = []
            signature_opt_nsarg_sig_args = []
            signature_mult_fn_sig_args = []
            signature_mult_nsarg_sig_args = []
            max_position = -1
            for sig_arg in signature["arguments"]:
                if "position" in sig_arg:
                    max_position = sig_arg["position"]
                    continue  # Skip position based signature arguments

                if (
                    sig_arg.get("optional", False) is True
                    and sig_arg.get("multiple", False) is False
                ):
                    if sig_arg["type"] in ["Function", "Modifier"]:
                        signature_opt_fn_sig_args.extend(sig_arg["values"])
                    elif sig_arg["type"] in ["NSArg", "StrArgNSArg"]:
                        signature_opt_nsarg_sig_args.extend(sig_arg["values"])
                elif sig_arg.get("multiple", False) is True:
                    if sig_arg["type"] in ["Function", "Modifier"]:
                        signature_mult_fn_sig_args.extend(sig_arg["values"])
                    elif sig_arg["type"] in ["NSArg", "StrArgNSArg"]:
                        signature_mult_nsarg_sig_args.extend(sig_arg["values"])

            # Remove signature non-multiple, optional arguments that are already in args list
            for idx, arg in enumerate(args):
                if idx <= max_position - 1:  # Skip positional arguments
                    continue
                if idx == arg_idx:  # Skip argument to be completed
                    continue
                log.debug(f"Remove Optional Args {arg}  {signature_opt_fn_sig_args}")

            opt_fn_sig_args.extend(signature_opt_fn_sig_args)
            opt_nsarg_sig_args.extend(signature_opt_nsarg_sig_args)
            mult_fn_sig_args.extend(signature_mult_fn_sig_args)
            mult_nsarg_sig_args.extend(signature_mult_nsarg_sig_args)

        function_list.extend(list(set(opt_fn_sig_args + mult_fn_sig_args)))
        entity_types.extend(list(set(opt_nsarg_sig_args + mult_nsarg_sig_args)))

    if function_list:
        log.debug(f"ArgComp - position-based Function list: {function_list}")
        fn_replace_list = function_completions(
            completion_text, bel_spec, function_list, bel_fmt, size
        )

    if entity_types:
        log.debug(f"ArgComp - position-based Entity types: {entity_types}")
        ns_arg_replace_list = nsarg_completions(
            completion_text,
            entity_types,
            bel_spec,
            namespace,
            species_id,
            bel_fmt,
            size,
        )

    replace_list = fn_replace_list + ns_arg_replace_list

    return replace_list


def add_completions(
    replace_list: list, belstr: str, replace_span: Span, completion_text: str
) -> List[Mapping[str, Any]]:
    """Create completions to return given replacement list

    Args:
        replace_list: list of completion replacement values
        belstr: BEL String
        replace_span: start, stop of belstr to replace
        completion_text: text to use for completion - used for creating highlight
    Returns:
        [{
            "replacement": replacement,
            "cursor_loc": cursor_loc,
            "highlight": highlight,
            "label": label,
        }]
    """

    completions = []

    for r in replace_list:

        # if '(' not in belstr:
        #     replacement = f'{r["replacement"]}()'
        #     cursor_loc = len(replacement) - 1  # inside parenthesis
        # elif r['type'] == 'Function' and replace_span[1] == len(belstr) - 1:

        if len(belstr) > 0:
            belstr_end = len(belstr) - 1
        else:
            belstr_end = 0

        log.debug(
            f'Replace list {r}  Replace_span {replace_span}  BELstr: {belstr} Len: {belstr_end} Test1 {r["type"] == "Function"}  Test2 {replace_span[1] + 1 == len(belstr)}'
        )

        # Put a space between comma and following function arg
        if (
            r["type"] == "Function"
            and replace_span[0] > 0
            and belstr[replace_span[0] - 1] == ","
        ):
            log.debug("prior char is a comma")
            replacement = (
                belstr[0 : replace_span[0]]
                + " "
                + f"{r['replacement']}()"
                + belstr[replace_span[1] + 1 :]
            )
            cursor_loc = len(
                belstr[0 : replace_span[0]] + " " + f"{r['replacement']}()"
            )
        # Put a space between comman and following NSArg or StrArg
        elif replace_span[0] > 0 and belstr[replace_span[0] - 1] == ",":
            log.debug("prior char is a comma")
            replacement = (
                belstr[0 : replace_span[0]]
                + " "
                + r["replacement"]
                + belstr[replace_span[1] + 1 :]
            )
            cursor_loc = len(belstr[0 : replace_span[0]] + " " + r["replacement"])
        # Add function to end of belstr
        elif r["type"] == "Function" and replace_span[1] >= belstr_end:
            replacement = belstr[0 : replace_span[0]] + f"{r['replacement']}()"
            cursor_loc = len(replacement) - 1  # inside parenthesis
            log.debug(f"Replacement: {replacement}")
        # Insert replacement in beginning or middle of belstr
        else:
            replacement = (
                belstr[0 : replace_span[0]]
                + r["replacement"]
                + belstr[replace_span[1] + 1 :]
            )
            cursor_loc = len(
                belstr[0 : replace_span[0]] + r["replacement"]
            )  # move cursor just past replacement

        completions.append(
            {
                "replacement": replacement,
                "cursor_loc": cursor_loc,
                "highlight": r["highlight"],
                "label": r["label"],
            }
        )

    return completions


def get_completions(
    belstr: str,
    cursor_loc: int,
    bel_spec: BELSpec,
    bel_comp: str,
    bel_fmt: str,
    species_id: str,
    size: int,
):
    """Get BEL Assertion completions

    Args:

    Results:

    """

    ast, errors = pparse.get_ast_dict(belstr)

    spans = pparse.collect_spans(ast)

    completion_text = ""
    completions = []
    function_help = []

    log.debug(f"Cursor location BELstr: {belstr}  Cursor idx: {cursor_loc}")
    cursor_results = cursor(belstr, ast, cursor_loc)
    log.debug(f"Cursor results: {cursor_results}")

    if not cursor_results:
        log.debug("Cursor results is empty")
        return (completion_text, completions, function_help, spans)

    completion_text = cursor_results.get("completion_text", "")

    replace_span = cursor_results["replace_span"]
    namespace = cursor_results.get("namespace", None)

    if "parent_function" in cursor_results:
        parent_function = cursor_results["parent_function"]
        function_help = bel_specification.get_function_help(
            cursor_results["parent_function"], bel_spec
        )

        args = cursor_results.get("args", [])
        arg_idx = cursor_results.get("arg_idx")

        replace_list = arg_completions(
            completion_text,
            parent_function,
            args,
            arg_idx,
            bel_spec,
            bel_fmt,
            species_id,
            namespace,
            size,
        )
    elif cursor_results["type"] == "Function":
        function_list = None
        replace_list = function_completions(
            completion_text, bel_spec, function_list, bel_fmt, size
        )
    elif cursor_results["type"] == "Relation":
        replace_list = relation_completions(completion_text, bel_spec, bel_fmt, size)

    completions.extend(
        add_completions(replace_list, belstr, replace_span, completion_text)
    )

    return completion_text, completions, function_help, spans


def bel_completion(
    belstr: str,
    cursor_loc: int = -1,
    bel_version: str = default_bel_version,
    bel_comp: str = None,
    bel_fmt: str = "medium",
    species_id: str = None,
    size: int = 20,
) -> Mapping[str, Any]:
    """BEL Completion

    Args:
        belstr (str): BEL String to provide completion for
        cursor_loc (int): cursor location - default of -1 means end of string
        bel_version (str): BEL Language version to use for completion
        bel_comp (str): ['subject', 'object', 'full', None] - a nested statement has to be found in object or full statement
        bel_fmt (str): ['short', 'medium', 'long'] BEL function/relation format
        species_id (str): optional, species id is used to filter namespace values if applicable (e.g. Gene, RNA, ... entity_types)
        size: how many completions to return, defaults to 20

    Returns:
        Mapping[str, Any]:
            {
                'completions': completions,
                'function_help': function_help,
                'entity_spans': spans
            }
    """

    """
    Completion object: {
        completions: [
            {
                'replacement': <replacement text field string,
                'cursor_loc': <new cursor location>
                'highlight': <highlighted match>
                'label': <label for completion>
            },
        ],
        function_help: [{
            "function_summary": <template>,
            "argument_help": [<argument help>],
            "description": <desc>
        }],
        "entity_spans": {<span info>}
    }

    """
    bel_spec = bel_specification.get_specification(bel_version)

    belstrlen = len(belstr)
    if cursor_loc == -1:
        cursor_loc = belstrlen - 1
    elif cursor_loc >= belstrlen:
        cursor_loc = belstrlen - 1

    # with timy.Timer() as timer:
    #     (completion_text, completions, function_help, spans) = get_completions(belstr, cursor_loc, bel_spec, bel_comp, bel_fmt, species_id, size)

    (completion_text, completions, function_help, spans) = get_completions(
        belstr, cursor_loc, bel_spec, bel_comp, bel_fmt, species_id, size
    )

    return {
        "completion_text": completion_text,
        "completions": completions,
        "function_help": function_help,
        "entity_spans": spans,
    }


def main():

    bel_version = "2.0.0"
    bel_spec = bel_specification.get_specification(bel_version)

    belstr = "compositeAbundance(proteinAbundance(HGNC:FN1), proteinAbundance(HGNC:VEGFA)) increases translocation(proteinAbundance(HGNC:PRKCA), fromLoc(MESH:Intracellular Space), toLoc(MESH:Cell Membrane))"
    belstr = 'proteinAbundance(HGNC:VEGFA) increases (compositeAbundance(proteinAbundance(HGNC:ITGB1), proteinAbundance(HGNC:PRKCA, ma(kin))) increases biologicalProcess(GO:"cell-matrix adhesion"))'
    belstr = "complex(p(HGNC:EGFR))"
    # belstr = 'p(HGNC:AKT)'

    cursor_loc = 18
    completions = bel_completion(belstr, cursor_loc, bel_fmt="long")
    print("Completions:\n", json.dumps(completions, indent=4))

    quit()
    ast, errors = pparse.get_ast_dict(belstr)
    # print('AST:\n', json.dumps(ast, indent=4))
    # quit()
    cursor_loc = 8
    results = cursor(belstr, ast, cursor_loc)
    print(f"Cursor{cursor_loc}:\n", json.dumps(results, indent=4))
    quit()

    bel_comp = ""
    bel_fmt = "long"
    species_id = "TAX:9606"
    completions = get_completions(
        belstr, cursor_loc, functions, bel_spec, bel_comp, bel_fmt, species_id
    )
    print("Completions:\n", json.dumps(completions, indent=4))
    quit()

    completions = ns_completions("AKT", ["Protein"], "TAX:9606", 10)
    print("Completions:\n", json.dumps(completions, indent=4))
    quit()

    # completions = bel_completion('pa', bel_fmt='long')
    # print('DumpVar:\n', json.dumps(completions, indent=4))

    completions = bel_completion("path(", cursor_loc=1, bel_fmt="medium")
    print("DumpVar:\n", json.dumps(completions, indent=4))


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Completion notes
1. Won't complete inside of a quote.  If there are mis-matched quotes
   it will break completions
2. terms upstream of a '(' are functions/modifiers
3. commas separate arguments of upstream function


"""
import re
import copy
from typing import Mapping, Any, List, Tuple, MutableMapping, Optional

import logging
import logging.config
import structlog

import bel.lang.bel_specification as bel_specification
from bel.lang.ast import BELAst, Function, NSArg, StrArg

from bel.Config import config

# if config.get('logging', False):
#     logging.config.dictConfig(config.get('logging'))
# log = logging.getLogger(__name__)

log = structlog.getLogger(__name__)

start_arg_chars = ["(", ","]
end_arg_chars = [")", ","]

relations_pattern_middle = re.compile(r"\)\s+([a-zA-Z\=\-\>\|\:]+)\s+[\w\(]+")
relations_pattern_end = re.compile(r"\)\s+([a-zA-Z\=\-\>\|\:]+)\s*$")

Errors = List[
    Tuple[str, str, Optional[Tuple[int, int]]]
]  # (<"Error"|"Warning">, "message", (start_span, end_span))
Parsed = MutableMapping[str, Any]
AST = MutableMapping[str, Any]
CharLocs = Mapping[str, Any]

"""Parsed data structure example
{
   "(77, 104)": {
        "name": "proteinAbundance",
        "type": "Function",
        "span": [
            77,
            104
        ],
        "name_span": [
            77,
            92
        ],
        "parens_span": [
            93,
            104
        ],
        "function_level": "child",
        "args": [
            {
                "arg": "HGNC:ITGB1",
                "span": [
                    94,
                    103
                ],
                "type": "NSArg",
                "ns": "HGNC",
                "ns_span": [
                    94,
                    97
                ],
                "ns_val": "ITGB1",
                "ns_val_span": [
                    99,
                    103
                ]
            }
        ]
    },
    "(29, 38)": {
        "type": "Relation",
        "name": "increases",
        "span": [
            29,
            38
        ]
    },
    "(160, 169)": {
        "type": "Relation",
        "name": "increases",
        "span": [
            160,
            169
        ]
    },
    "(39, 214)": {
        "type": "Nested",
        "span": [
            39,
            214
        ]
    }
}
"""


def parse_chars(bels: list, errors: Errors) -> Tuple[CharLocs, Errors]:
    """Scan BEL string to map parens, quotes, commas

    Args:
        bels: bel string as an array of characters
        errors: list of error tuples ('<type>', '<msg>')

    Returns:
        (char_locs, errors): character locations and errors
    """
    pstack, qstack, nested_pstack = [], [], []
    parens, nested_parens, quotes, commas = {}, {}, {}, {}

    notquoted_flag = True

    for i, c in enumerate(bels):
        prior_char = i - 1
        # print('BEL', prior_char, b[prior_char])

        # Find starting quote
        if c == '"' and bels[prior_char] != "\\" and len(qstack) == 0:
            qstack.append(i)
            notquoted_flag = False

        # Find closing quote
        elif c == '"' and bels[prior_char] != "\\":
            quotes[qstack.pop()] = i
            notquoted_flag = True

        # Find all escaped quotes outside of quoted string
        elif c == '"' and bels[prior_char] == "\\" and len(qstack) == 0:
            errors.append(
                (
                    "ERROR",
                    f"Escaped quote outside of quoted string at location: {i - 1}",
                    (i - 1, i - 1),
                )
            )

        # Find all nested object opening parens
        elif notquoted_flag and c == "(" and bels[prior_char] == " ":
            if len(nested_pstack) > 1:
                errors.append(
                    (
                        "ERROR",
                        f"More than one nested parenthesis or left parenthesis following a space character",
                        (i, i),
                    )
                )

            nested_pstack.append(i)

        # Find all opening parens
        elif notquoted_flag and c == "(" and bels[prior_char] not in ["\\"]:
            pstack.append(i)

        # Find all closing parens
        elif notquoted_flag and c == ")" and bels[prior_char] != "\\":
            if len(pstack):
                if len(pstack) > 1:
                    parens[pstack.pop()] = (i, "child")
                else:
                    parens[pstack.pop()] = (i, "top")
            elif len(nested_pstack):
                nested_parens[nested_pstack.pop()] = (i, "top")
            else:
                errors.append(
                    (
                        "ERROR",
                        f"Missing left parenthesis for right parenthesis at location {i}",
                        (i, i),
                    )
                )
        # Find comma outside of quoted string
        elif notquoted_flag and c == "," and len(qstack) == 0:
            sparen = pstack[-1]
            if sparen not in commas:
                commas[sparen] = [i]
            else:
                commas[sparen].append(i)

    while len(pstack):
        errors.append(
            (
                "ERROR",
                f"Missing right parenthesis for left parenthesis at location {pstack[-1]}",
                (pstack[-1], pstack[-1]),
            )
        )
        if len(pstack) > 1:
            parens[pstack.pop()] = (-1, "child")
        else:
            parens[pstack.pop()] = (-1, "top")

    while len(nested_pstack):
        errors.append(
            (
                "ERROR",
                f"Missing right parenthesis for nested object left parenthesis at location {nested_pstack[-1]}",
                (nested_pstack[-1], nested_pstack[-1]),
            )
        )
        nested_parens[nested_pstack.pop()] = (-1, "top")

    if len(qstack):
        missing_quote = qstack.pop()
        errors.append(
            (
                "ERROR",
                f"Missing right quote for left quote at location {missing_quote}",
                (missing_quote, missing_quote),
            )
        )

    return (
        {
            "parens": parens,
            "nested_parens": nested_parens,
            "quotes": quotes,
            "commas": commas,
        },
        errors,
    )


def parse_functions(
    bels: list, char_locs: CharLocs, parsed: Parsed, errors: Errors
) -> Tuple[Parsed, Errors]:
    """Parse functions from BEL using paren, comma, quote character locations

    Args:
        bels: BEL string as list of chars
        char_locs: paren, comma, quote character locations
        errors: Any error messages generated during the parse

    Returns:
        (functions, errors): function names and locations and error messages
    """
    parens = char_locs["parens"]

    # Handle partial top-level function name
    if not parens:
        bels_len = len(bels) - 1
        span = (0, bels_len)
        parsed[span] = {
            "name": "".join(bels),
            "type": "Function",
            "span": span,
            "name_span": (span),
            "function_level": "top",
        }
        return parsed, errors

    for sp in sorted(parens):  # sp = starting paren, ep = ending_paren
        ep, function_level = parens[sp]

        # Functions can't have a space between function name and left paren
        if bels[sp - 1] == " ":
            continue

        # look in front of start paren for function name
        for i in range(sp - 1, 0, -1):
            if bels[i] in [" ", ",", "("]:  # function name upstream boundary chars
                if i < sp - 1:
                    if ep == -1:
                        span = (i + 1, len(bels) - 1)
                    else:
                        span = (i + 1, ep)

                    parsed[span] = {
                        "name": "".join(bels[i + 1 : sp]),
                        "type": "Function",
                        "span": span,
                        "name_span": (i + 1, sp - 1),
                        "parens_span": (sp, ep),
                        "function_level": function_level,
                    }
                break
        else:
            if ep == -1:
                span = (0, len(bels) - 1)
            else:
                span = (0, ep)

            parsed[span] = {
                "name": "".join(bels[0:sp]),
                "type": "Function",
                "span": span,
                "name_span": (0, sp - 1),
                "parens_span": (sp, ep),
                "function_level": function_level,
            }

    return parsed, errors


def parse_args(
    bels: list, char_locs: CharLocs, parsed: Parsed, errors: Errors
) -> Tuple[Parsed, Errors]:
    """Parse arguments from functions

    Args:
        bels: BEL string as list of chars
        char_locs: char locations for parens, commas and quotes
        parsed: function locations
        errors: error messages

    Returns:
        (functions, errors): function and arg locations plus error messages
    """

    commas = char_locs["commas"]

    # Process each span key in parsed from beginning
    for span in parsed:
        if parsed[span]["type"] != "Function" or "parens_span" not in parsed[span]:
            continue  # Skip if not argument-less
        sp, ep = parsed[span]["parens_span"]

        # calculate args_end position
        if ep == -1:  # supports bel completion
            args_end = len(bels) - 1  # 1
        else:
            args_end = ep - 1  # 1

        # Parse arguments
        args = []
        arg_start = sp + 1
        each_arg_end_list = sorted([end - 1 for end in commas.get(sp, [])] + [args_end])
        for arg_end in each_arg_end_list:
            # log.debug(f'Arg_start: {arg_start}  Arg_end: {arg_end}')

            # Skip blanks at beginning of argument
            while arg_start < args_end and bels[arg_start] == " ":
                arg_start += 1

            # Trim arg_end (e.g. HGNC:AKT1  , HGNC:EGF) - if there are spaces before comma
            trimmed_arg_end = arg_end
            while trimmed_arg_end > arg_start and bels[trimmed_arg_end] == " ":
                trimmed_arg_end -= 1

            if trimmed_arg_end < arg_start:
                trimmed_arg_end = arg_start

            arg = "".join(bels[arg_start : trimmed_arg_end + 1])

            # log.debug(f'Adding arg to args: {arg_start} {trimmed_arg_end}')
            args.append({"arg": arg, "span": (arg_start, trimmed_arg_end)})
            arg_start = arg_end + 2

        parsed[span]["args"] = args

    return parsed, errors


def arg_types(parsed: Parsed, errors: Errors) -> Tuple[Parsed, Errors]:
    """Add argument types to parsed function data structure

    Args:
        parsed: function and arg locations in BEL string
        errors: error messages

    Returns:
        (parsed, errors): parsed, arguments with arg types plus error messages
    """

    func_pattern = re.compile(r"\s*[a-zA-Z]+\(")
    nsarg_pattern = re.compile(r"^\s*([A-Z]+):(.*?)\s*$")

    for span in parsed:
        if parsed[span]["type"] != "Function" or "parens_span" not in parsed[span]:
            continue

        for i, arg in enumerate(parsed[span]["args"]):
            nsarg_matches = nsarg_pattern.match(arg["arg"])
            if func_pattern.match(arg["arg"]):
                parsed[span]["args"][i].update({"type": "Function"})
            elif nsarg_matches:
                (start, end) = arg["span"]
                ns = nsarg_matches.group(1)
                ns_val = nsarg_matches.group(2)
                ns_span = nsarg_matches.span(1)
                ns_span = (ns_span[0] + start, ns_span[1] + start - 1)
                ns_val_span = nsarg_matches.span(2)
                ns_val_span = (ns_val_span[0] + start, ns_val_span[1] + start - 1)

                parsed[span]["args"][i].update(
                    {
                        "type": "NSArg",
                        "ns": ns,
                        "ns_span": ns_span,
                        "ns_val": ns_val,
                        "ns_val_span": ns_val_span,
                    }
                )
            else:
                parsed[span]["args"][i].update({"type": "StrArg"})

    return parsed, errors


def parse_relations(
    belstr: str, char_locs: CharLocs, parsed: Parsed, errors: Errors
) -> Tuple[Parsed, Errors]:
    """Parse relations from BEL string

    Args:
        belstr: BEL string as one single string (not list of chars)
        char_locs: paren, comma and quote char locations
        parsed: data structure for parsed functions, relations, nested
        errors: error messages

    Returns:
        (parsed, errors):
    """
    quotes = char_locs["quotes"]
    quoted_range = set([i for start, end in quotes.items() for i in range(start, end)])

    for match in relations_pattern_middle.finditer(belstr):
        (start, end) = match.span(1)
        # log.debug(f'Relation-middle {match}')
        end = end - 1  # adjust end to match actual end character index
        if start != end:
            test_range = set(range(start, end))
        else:
            test_range = set(start)

        # Skip if relation overlaps with quoted string
        if test_range.intersection(quoted_range):
            continue

        span_key = (start, end)
        parsed[span_key] = {
            "type": "Relation",
            "name": match.group(1),
            "span": (start, end),
        }

    for match in relations_pattern_end.finditer(belstr):
        (start, end) = match.span(1)
        log.debug(f"Relation-end {match}")
        end = end - 1  # adjust end to match actual end character index
        if start != end:
            test_range = set(range(start, end))
        else:
            test_range = set(start)

        # Skip if relation overlaps with quoted string
        if test_range.intersection(quoted_range):
            continue

        span_key = (start, end)
        parsed[span_key] = {
            "type": "Relation",
            "name": match.group(1),
            "span": (start, end),
        }

    return parsed, errors


def parse_nested(
    bels: list, char_locs: CharLocs, parsed: Parsed, errors: Errors
) -> Tuple[Parsed, Errors]:
    """ Parse nested BEL object """

    for sp in char_locs[
        "nested_parens"
    ]:  # sp = start parenthesis, ep = end parenthesis
        ep, level = char_locs["nested_parens"][sp]
        if ep == -1:
            ep = len(bels) + 1
        parsed[(sp, ep)] = {"type": "Nested", "span": (sp, ep)}

    return parsed, errors


def dump_json(d: dict) -> None:
    """Dump json when using tuples for dictionary keys

    Have to convert tuples to strings to dump out as json
    """

    import json

    k = d.keys()
    v = d.values()
    k1 = [str(i) for i in k]

    return json.dumps(dict(zip(*[k1, v])), indent=4)


def collect_spans(ast: AST) -> List[Tuple[str, Tuple[int, int]]]:
    """Collect flattened list of spans of BEL syntax types

    Provide simple list of BEL syntax type spans for highlighting.
    Function names, NSargs, NS prefix, NS value and StrArgs will be
    tagged.

    Args:
        ast: AST of BEL assertion

    Returns:
        List[Tuple[str, Tuple[int, int]]]: list of span objects (<type>, (<start>, <end>))
    """

    spans = []

    if ast.get("subject", False):
        spans.extend(collect_spans(ast["subject"]))

    if ast.get("object", False):
        spans.extend(collect_spans(ast["object"]))

    if ast.get("nested", False):
        spans.extend(collect_spans(ast["nested"]))

    if ast.get("function", False):
        log.debug(f"Processing function")
        spans.append(("Function", ast["function"]["name_span"]))
        log.debug(f"Spans: {spans}")

    if ast.get("args", False):
        for idx, arg in enumerate(ast["args"]):
            log.debug(f"Arg  {arg}")

            if arg.get("function", False):
                log.debug(f"Recursing on arg function")
                results = collect_spans(arg)
                log.debug(f"Results {results}")
                spans.extend(results)  # Recurse arg function
            elif arg.get("nsarg", False):
                log.debug(f"Processing NSArg   Arg {arg}")
                spans.append(("NSArg", arg["span"]))
                spans.append(("NSPrefix", arg["nsarg"]["ns_span"]))
                spans.append(("NSVal", arg["nsarg"]["ns_val_span"]))
            elif arg["type"] == "StrArg":
                spans.append(("StrArg", arg["span"]))

    log.debug(f"Spans: {spans}")
    return spans

    # max_idx = 0
    # for key in parsed:
    #     if parsed[key]['type'] == 'Function':
    #         spans.append(('Function', parsed[key]['name_span']))
    #         if parsed[key]['name_span'][1] > max_idx:
    #             max_idx = parsed[key]['name_span'][1]

    #         for arg in parsed[key]['args']:
    #             if arg['type'] == 'NSArg':
    #                 spans.append(('NSArg', arg['span']))
    #                 spans.append(('Prefix', arg['ns_span']))
    #                 spans.append(('Value', arg['ns_val_span']))
    #                 if arg['ns_val_span'][1] > max_idx:
    #                     max_idx = arg['ns_val_span'][1]
    #             if arg['type'] == 'StrArg':
    #                 spans.append(('StrArg', arg['span']))
    #                 if arg['span'][1] > max_idx:
    #                     max_idx = arg['span'][1]
    #     elif parsed[key]['type'] == 'Relation':
    #         spans.append(('Relation', parsed[key]['span']))

    #     elif parsed[key]['type'] == 'Nested':
    #         spans.append(('Nested', parsed[key]['span']))


def print_spans(spans, max_idx: int) -> None:
    """Quick test to show how character spans match original BEL String

    Mostly for debugging purposes
    """

    bel_spans = [" "] * (max_idx + 3)
    for val, span in spans:
        if val in ["Nested", "NSArg"]:
            continue
        for i in range(span[0], span[1] + 1):
            bel_spans[i] = val[0]

    # print(''.join(bel_spans))

    # Add second layer for Nested Objects if available
    bel_spans = [" "] * (max_idx + 3)
    for val, span in spans:
        if val not in ["Nested"]:
            continue
        for i in range(span[0], span[1] + 1):
            bel_spans[i] = val[0]

    # print(''.join(bel_spans))


def parsed_function_to_ast(parsed: Parsed, parsed_key):
    """Create AST for top-level functions"""

    sub = parsed[parsed_key]

    subtree = {
        "type": "Function",
        "span": sub["span"],
        "function": {
            "name": sub["name"],
            "name_span": sub["name_span"],
            "parens_span": sub.get("parens_span", []),
        },
    }

    args = []
    for arg in parsed[parsed_key].get("args", []):

        # pdb.set_trace()

        if arg["type"] == "Function":
            args.append(parsed_function_to_ast(parsed, arg["span"]))
        elif arg["type"] == "NSArg":
            args.append(
                {
                    "arg": arg["arg"],
                    "type": arg["type"],
                    "span": arg["span"],
                    "nsarg": {
                        "ns": arg["ns"],
                        "ns_val": arg["ns_val"],
                        "ns_span": arg["ns_span"],
                        "ns_val_span": arg["ns_val_span"],
                    },
                }
            )
        elif arg["type"] == "StrArg":
            args.append({"arg": arg["arg"], "type": arg["type"], "span": arg["span"]})

    subtree["args"] = copy.deepcopy(args)

    return subtree


def parsed_top_level_errors(parsed, errors, component_type: str = "") -> Errors:
    """Check full parse for errors

    Args:
        parsed:
        errors:
        component_type: Empty string or 'subject' or 'object' to indicate that we
            are parsing the subject or object field input
    """

    # Error check
    fn_cnt = 0
    rel_cnt = 0
    nested_cnt = 0
    for key in parsed:
        if parsed[key]["type"] == "Function":
            fn_cnt += 1
        if parsed[key]["type"] == "Relation":
            rel_cnt += 1
        if parsed[key]["type"] == "Nested":
            nested_cnt += 1

    if not component_type:
        if nested_cnt > 1:
            errors.append(
                (
                    "Error",
                    "Too many nested objects - can only have one per BEL Assertion",
                )
            )

        if nested_cnt:
            if rel_cnt > 2:
                errors.append(
                    (
                        "Error",
                        "Too many relations - can only have two in a nested BEL Assertion",
                    )
                )
            elif fn_cnt > 4:
                errors.append(("Error", "Too many BEL subject and object candidates"))

        else:
            if rel_cnt > 1:
                errors.append(
                    (
                        "Error",
                        "Too many relations - can only have one in a BEL Assertion",
                    )
                )
            elif fn_cnt > 2:
                errors.append(("Error", "Too many BEL subject and object candidates"))

    elif component_type == "subject":
        if rel_cnt > 0:
            errors.append(
                ("Error", "Too many relations - cannot have any in a BEL Subject")
            )
        elif fn_cnt > 1:
            errors.append(
                ("Error", "Too many BEL subject candidates - can only have one")
            )

    elif component_type == "object":
        if nested_cnt:
            if rel_cnt > 1:
                errors.append(
                    (
                        "Error",
                        "Too many relations - can only have one in a nested BEL object",
                    )
                )
            elif fn_cnt > 2:
                errors.append(
                    (
                        "Error",
                        "Too many BEL subject and object candidates in a nested BEL object",
                    )
                )
        else:
            if rel_cnt > 0:
                errors.append(
                    ("Error", "Too many relations - cannot have any in a BEL Subject")
                )
            elif fn_cnt > 1:
                errors.append(
                    ("Error", "Too many BEL subject candidates - can only have one")
                )

    return errors


def parsed_to_ast(parsed: Parsed, errors: Errors, component_type: str = ""):
    """Convert parsed data struct to AST dictionary

    Args:
        parsed:
        errors:
        component_type: Empty string or 'subject' or 'object' to indicate that we
            are parsing the subject or object field input
    """

    ast = {}
    sorted_keys = sorted(parsed.keys())

    # Setup top-level tree
    for key in sorted_keys:
        if parsed[key]["type"] == "Nested":
            nested_component_stack = ["subject", "object"]

    if component_type:
        component_stack = [component_type]
    else:
        component_stack = ["subject", "object"]

    for key in sorted_keys:
        if parsed[key]["type"] == "Function" and parsed[key]["function_level"] == "top":
            ast[component_stack.pop(0)] = parsed_function_to_ast(parsed, key)
        elif parsed[key]["type"] == "Relation" and "relation" not in ast:
            ast["relation"] = {
                "name": parsed[key]["name"],
                "type": "Relation",
                "span": key,
            }
        elif parsed[key]["type"] == "Nested":
            ast["nested"] = {}
            for nested_key in sorted_keys:
                if nested_key <= key:
                    continue

                if (
                    parsed[nested_key]["type"] == "Function"
                    and parsed[nested_key]["function_level"] == "top"
                ):
                    ast["nested"][
                        nested_component_stack.pop(0)
                    ] = parsed_function_to_ast(parsed, nested_key)
                elif (
                    parsed[nested_key]["type"] == "Relation"
                    and "relation" not in ast["nested"]
                ):
                    ast["nested"]["relation"] = {
                        "name": parsed[nested_key]["name"],
                        "type": "Relation",
                        "span": parsed[nested_key]["span"],
                    }

            return ast, errors

    return ast, errors


def get_ast_dict(belstr, component_type: str = ""):
    """Convert BEL string to AST dictionary

    Args:
        belstr: BEL string
        component_type: Empty string or 'subject' or 'object' to indicate that we
            are parsing the subject or object field input
    """

    errors = []
    parsed = {}
    bels = list(belstr)
    char_locs, errors = parse_chars(bels, errors)
    parsed, errors = parse_functions(belstr, char_locs, parsed, errors)
    parsed, errors = parse_args(bels, char_locs, parsed, errors)
    parsed, errors = arg_types(parsed, errors)
    parsed, errors = parse_relations(belstr, char_locs, parsed, errors)
    parsed, errors = parse_nested(bels, char_locs, parsed, errors)
    errors = parsed_top_level_errors(parsed, errors)

    ast, errors = parsed_to_ast(parsed, errors, component_type=component_type)

    return ast, errors


def get_ast_obj(belstr, bel_version, component_type: str = ""):
    """Convert AST partialparse dict to BELAst"""

    ast_dict, errors = get_ast_dict(belstr, component_type)

    spec = bel_specification.get_specification(bel_version)

    subj = ast_dict["subject"]
    subj_ast = add_ast_fn(subj, spec)

    relation = None
    obj = None

    if "relation" in ast_dict:
        relation = ast_dict["relation"]["name"]

    if "object" in ast_dict:
        obj = ast_dict["object"]
        obj_ast = add_ast_fn(obj, spec)

        return BELAst(subj_ast, relation, obj_ast, spec)
    elif "nested" in ast_dict:
        nested_subj = ast_dict["nested"]["subject"]
        nested_subj_ast = add_ast_fn(nested_subj, spec)
        nested_relation = ast_dict["nested"]["relation"]["name"]
        nested_obj = ast_dict["nested"]["object"]
        nested_obj_ast = add_ast_fn(nested_obj, spec)

        return BELAst(
            subj_ast,
            relation,
            BELAst(nested_subj_ast, nested_relation, nested_obj_ast, spec),
            spec,
        )

    return BELAst(subj_ast, None, None, spec)


def add_ast_fn(d, spec, parent_function=None):
    """Convert dict AST to object AST Function

    Args:
        ast_fn: AST object Function
        d: AST as dictionary
        spec: BEL Specification

    Return:
        ast_fn
    """

    if d["type"] == "Function":
        ast_fn = Function(d["function"]["name"], spec, parent_function=parent_function)
        for arg in d["args"]:
            if arg["type"] == "Function":
                ast_fn.add_argument(add_ast_fn(arg, spec, parent_function=ast_fn))
            elif arg["type"] == "NSArg":
                ast_fn.add_argument(
                    NSArg(arg["nsarg"]["ns"], arg["nsarg"]["ns_val"], ast_fn)
                )
            elif arg["type"] == "StrArg":
                ast_fn.add_argument(StrArg(arg["arg"], ast_fn))
    return ast_fn


def main():

    import json

    belstr = 'activity(proteinAbundance(SFAM:"GSK3 "Family"), molecularActivity(DEFAULT:kin))'
    belstr = 'proteinAbundance(HGNC:VHL) increases (proteinAbundance(HGNC:TNF) increases biologicalProcess(GOBP:"cell death"))'
    belstr = (
        "complexAbundance(proteinAbundance(HGNC:VHL), proteinAbundance(HGNC:PRKCZ))"
    )
    # made up (added the 20 in the pmod)
    belstr = 'activity(proteinAbundance(SFAM:"PRKA Family"), molecularActivity(DEF:kin)) directlyIncreases proteinAbundance(SFAM:"PDE4 Long Family", proteinModification(Ph, S, 20))'
    belstr = 'proteinAbundance(HGNC:VEGFA) increases (compositeAbundance(proteinAbundance(HGNC:ITGB1), proteinAbundance(HGNC:PRKCA, ma(kin))) increases biologicalProcess(GO:"cell-matrix adhesion"))'
    belstr = "complex(p(HGNC:AKT1))"
    # belstr = 'p(fus(HGNC:EGF, 20, '
    # belstr = 'pa'

    ast, errors = get_ast_dict(belstr)

    spans = collect_spans(ast)

    print("\n\nBELStr", belstr)
    print("Spans:\n", json.dumps(spans, indent=4))

    # print('AST:\n', json.dumps(ast, indent=4))


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This file contains functions to enhance, retrieve and return BEL object specifications.

NOTE !!!!  Have to BEL Spec JSON files if the BEL Spec JSON format is changed here !!!
           Run `make belspec_json`

"""

import glob
import os
import re
import copy
import requests
import sys
import yaml
import datetime
import json
import itertools
from typing import Mapping, List, Any
import jinja2
import tatsu
import importlib

from bel.Config import config

import structlog

log = structlog.getLogger(__name__)

# Custom Typing definitions
BELSpec = Mapping[str, Any]


"""
Keys available in enhanced spec_dict:

- version
- notes
- admin
    - 'version_underscored'
    - 'parser_fn'
- relations
    - 'list'
    - 'list_short'
    - 'list_long'
    - 'to_short'
    - 'to_long'
    - 'info'
- functions
    - 'argument_types'
    - 'entity_types'
    - 'info'
    - 'signatures'
    - 'list'
    - 'list_short'
    - 'list_long'
    - 'primary'
        - 'list_short'
        - 'list_long'
    - 'modifier'
        - 'list_short'
        - 'list_long'
    - 'to_short'
    - 'to_long'
- namespaces
    - 'Activity'
        - info
        - list
        - list_short
        - list_long
        - to_short
        - to_long
    - 'ProteinModification'
        - info
        - list
        - list_short
        - list_long
        - to_short
        - to_long
    - 'AminoAcid'
        - info
        - list
        - list_short
        - list_long
        - to_short
        - to_long


- computed_signatures
"""


def get_specification(version: str) -> Mapping[str, Any]:
    """Get BEL Specification

    The json file this depends on is generated by belspec_yaml2json as
    part of the update_specifications function

    Args:
        version: e.g. 2.0.0 where the filename
    """

    spec_dir = config["bel"]["lang"]["specifications"]
    spec_dict = {}

    bel_versions = get_bel_versions()
    if version not in bel_versions:
        log.error("Cannot get unknown version BEL specification")
        return {"error": "unknown version of BEL"}

    # use this variable to find our parser file since periods aren't recommended in python module names
    version_underscored = version.replace(".", "_")

    json_fn = f"{spec_dir}/bel_v{version_underscored}.json"

    with open(json_fn, "r") as f:
        spec_dict = json.load(f)

    return spec_dict


def get_bel_versions() -> List[str]:
    """Get BEL Language versions supported

    Get the list of all BEL Language versions supported.  The file this depends
    on is generated by belspec_yaml2json and is kept up to date using
    `make update_ebnf` or `make update_parsers`.  You can also run `belspec_yaml2json`
    directly as it's added as a command by pip install.

    Returns:
        List[str]: list of versions
    """

    spec_dir = config["bel"]["lang"]["specifications"]

    fn = f"{spec_dir}/versions.json"
    with open(fn, "r") as f:
        versions = json.load(f)

    return versions


def update_specifications(force: bool = False):
    """Update BEL specifications

    Collect BEL specifications from Github BELBio BEL Specification folder
    and store in local directory specified in belbio_conf.yaml

    Process all BEL Specifications in YAML into an enhanced JSON version
    and capture all BEL versions in a separate file for quick access.
    """

    spec_dir = config["bel"]["lang"]["specifications"]
    if not os.path.isdir(spec_dir):
        os.mkdir(spec_dir)

    log.info(f"Updating BEL Specifications - stored in {spec_dir}")

    # Collect new specifications from Git repository
    if config["bel"]["lang"]["specification_github_repo"]:
        github_belspec_files(spec_dir, force=force)

    # Ensure that files use 'yaml' extension
    files = glob.glob(f"{spec_dir}/*.yml")
    for fn in files:
        new_fn = fn.replace("yml", "yaml")
        os.rename(fn, new_fn)

    # Convert YAML to enhanced JSON
    files = glob.glob(f"{spec_dir}/bel_*.yaml")
    versions = {}
    for fn in files:
        filename = os.path.basename(fn)

        check_version = filename.replace("bel_v", "").replace(".yaml", "").replace("_", ".")

        json_fn = fn.replace(".yaml", ".json")
        version = belspec_yaml2json(fn, json_fn)

        if version != check_version:
            log.error(
                f"Version mis-match for {fn} - fn version: {check_version} version: {version}"
            )
        versions[version] = filename

    with open(f"{spec_dir}/versions.json", "w") as f:
        json.dump(list(set(versions)), f, indent=4)

    # Convert YAML file to EBNF and then parser module

    create_ebnf_parser(files)


def github_belspec_files(spec_dir, force: bool = False):
    """Get belspec files from Github repo


    Args:
        spec_dir: directory to store the BEL Specification and derived files
        force: force update of BEL Specifications from Github - skipped if local files less than 1 day old
    """

    if not force:
        dtnow = datetime.datetime.utcnow()
        delta = datetime.timedelta(1)
        yesterday = dtnow - delta

        for fn in glob.glob(f"{spec_dir}/bel*yaml"):
            if datetime.datetime.fromtimestamp(os.path.getmtime(fn)) > yesterday:
                log.info("Skipping BEL Specification update - specs less than 1 day old")
                return

    repo_url = "https://api.github.com/repos/belbio/bel_specifications/contents/specifications"
    params = {}
    github_access_token = os.getenv("GITHUB_ACCESS_TOKEN", "")
    if github_access_token:
        params = {"access_token": github_access_token}

    r = requests.get(repo_url, params=params)
    if r.status_code == 200:
        results = r.json()
        for f in results:
            url = f["download_url"]
            fn = os.path.basename(url)

            if "yaml" not in fn and "yml" in fn:
                fn = fn.replace("yml", "yaml")

            r = requests.get(url, params=params, allow_redirects=True)
            if r.status_code == 200:
                open(f"{spec_dir}/{fn}", "wb").write(r.content)
            else:
                sys.exit(
                    f"Could not get BEL Spec file {url} from Github -- Status: {r.status_code}  Msg: {r.content}"
                )
    else:
        sys.exit(
            f"Could not get BEL Spec directory listing from Github -- Status: {r.status_code}  Msg: {r.content}"
        )


def belspec_yaml2json(yaml_fn: str, json_fn: str) -> str:
    """Enhance BEL specification and save as JSON file

    Load all BEL Specification YAML files and convert to JSON files
    after enhancing them.  Also create a bel_versions.json file with
    all available BEL versions for fast loading.

    Args:
        yaml_fn: original YAML version of BEL Spec
        json_fn: enhanced JSON version of BEL Spec
    Returns:
        str: version of BEL Spec
    """

    try:
        spec_dict = yaml.load(open(yaml_fn, "r").read(), Loader=yaml.SafeLoader)

        # admin-related keys
        spec_dict["admin"] = {}
        spec_dict["admin"]["version_underscored"] = spec_dict["version"].replace(".", "_")
        spec_dict["admin"]["parser_fn"] = yaml_fn.replace(".yaml", "_parser.py")

        # add relation keys list, to_short, to_long
        add_relations(spec_dict)
        # add function keys list, to_short, to_long
        add_functions(spec_dict)
        # add namespace keys list, list_short, list_long, to_short, to_long
        add_namespaces(spec_dict)

        enhance_function_signatures(spec_dict)

        add_function_signature_help(spec_dict)

        with open(json_fn, "w") as f:
            json.dump(spec_dict, f)

    except Exception as e:
        log.error(
            f"Warning: BEL Specification {yaml_fn} could not be read. Cannot proceed -- error {str(e)}"
        )
        print("DumpVar:\n", json.dumps(spec_dict, indent=4))
        sys.exit()

    return spec_dict["version"]


def add_function_signature_help(spec_dict: dict) -> dict:
    """Add function signature help

    Simplify the function signatures for presentation to BEL Editor users
    """
    for f in spec_dict["functions"]["signatures"]:
        for argset_idx, argset in enumerate(spec_dict["functions"]["signatures"][f]["signatures"]):
            args_summary = ""
            args_list = []
            arg_idx = 0
            for arg_idx, arg in enumerate(
                spec_dict["functions"]["signatures"][f]["signatures"][argset_idx]["arguments"]
            ):
                if arg["type"] in ["Function", "Modifier"]:
                    vals = [
                        spec_dict["functions"]["to_short"].get(
                            val, spec_dict["functions"]["to_short"].get(val)
                        )
                        for val in arg["values"]
                    ]
                    args_summary += "|".join(vals) + "()"
                    arg_idx += 1

                    if arg.get("optional", False) and arg.get("multiple", False) is False:
                        args_summary += "?"
                        text = f'Zero or one of each function(s): {", ".join([val for val in arg["values"]])}'
                    elif arg.get("optional", False):
                        args_summary += "*"
                        text = f'Zero or more of each function(s): {", ".join([val for val in arg["values"]])}'
                    else:
                        text = f'One of following function(s): {", ".join([val for val in arg["values"]])}'

                elif arg["type"] in ["NSArg", "StrArg", "StrArgNSArg"]:
                    args_summary += f'{arg["type"]}'
                    if arg.get("optional", False) and arg.get("multiple", False) is False:
                        args_summary += "?"
                        if arg["type"] in ["NSArg"]:
                            text = f'Zero or one namespace argument of following type(s): {", ".join([val for val in arg["values"]])}'
                        elif arg["type"] == "StrArgNSArg":
                            text = f'Zero or one amespace argument or default namespace argument (without prefix) of following type(s): {", ".join([val for val in arg["values"]])}'
                        else:
                            text = f'Zero or one string argument of following type(s): {", ".join([val for val in arg["values"]])}'
                    elif arg.get("optional", False):
                        args_summary += "*"
                        if arg["type"] in ["NSArg"]:
                            text = f'Zero or more namespace arguments of following type(s): {", ".join([val for val in arg["values"]])}'
                        elif arg["type"] == "StrArgNSArg":
                            text = f'Zero or more namespace arguments or default namespace arguments (without prefix) of following type(s): {", ".join([val for val in arg["values"]])}'
                        else:
                            text = f'Zero or more of string arguments of following type(s): {", ".join([val for val in arg["values"]])}'
                    else:
                        if arg["type"] in ["NSArg"]:
                            text = f'Namespace argument of following type(s): {", ".join([val for val in arg["values"]])}'
                        elif arg["type"] == "StrArgNSArg":
                            text = f'Namespace argument or default namespace argument (without prefix) of following type(s): {", ".join([val for val in arg["values"]])}'
                        else:
                            text = f'String argument of following type(s): {", ".join([val for val in arg["values"]])}'

                args_summary += ", "
                args_list.append(text)

            args_summary = re.sub(", $", "", args_summary)
            spec_dict["functions"]["signatures"][f]["signatures"][argset_idx][
                "argument_summary"
            ] = f"{f}({args_summary})"
            spec_dict["functions"]["signatures"][f]["signatures"][argset_idx][
                "argument_help_listing"
            ] = args_list

            # print(f'{f}({args_summary})')
            # print(args_list)

    return spec_dict


def add_relations(spec_dict: Mapping[str, Any]) -> Mapping[str, Any]:
    """Add relation keys to spec_dict

    Args:
        spec_dict (Mapping[str, Any]): bel specification dictionary

    Returns:
        Mapping[str, Any]: bel specification dictionary with added relation keys
    """

    # Class 'Mapping' does not define '__setitem__', so the '[]' operator cannot be used on its instances
    spec_dict["relations"]["list"] = []
    spec_dict["relations"]["list_short"] = []
    spec_dict["relations"]["list_long"] = []
    spec_dict["relations"]["to_short"] = {}
    spec_dict["relations"]["to_long"] = {}

    for relation_name in spec_dict["relations"]["info"]:

        abbreviated_name = spec_dict["relations"]["info"][relation_name]["abbreviation"]
        spec_dict["relations"]["list"].extend((relation_name, abbreviated_name))
        spec_dict["relations"]["list_long"].append(relation_name)
        spec_dict["relations"]["list_short"].append(abbreviated_name)

        spec_dict["relations"]["to_short"][relation_name] = abbreviated_name
        spec_dict["relations"]["to_short"][abbreviated_name] = abbreviated_name

        spec_dict["relations"]["to_long"][abbreviated_name] = relation_name
        spec_dict["relations"]["to_long"][relation_name] = relation_name

    return spec_dict


def add_functions(spec_dict: Mapping[str, Any]) -> Mapping[str, Any]:
    """Add function keys to spec_dict

    Args:
        spec_dict (Mapping[str, Any]): bel specification dictionary

    Returns:
        Mapping[str, Any]: bel specification dictionary with added function keys
    """

    # Class 'Mapping' does not define '__setitem__', so the '[]' operator cannot be used on its instances
    spec_dict["functions"]["list"] = []
    spec_dict["functions"]["list_long"] = []
    spec_dict["functions"]["list_short"] = []

    spec_dict["functions"]["primary"] = {}
    spec_dict["functions"]["primary"]["list_long"] = []
    spec_dict["functions"]["primary"]["list_short"] = []

    spec_dict["functions"]["modifier"] = {}
    spec_dict["functions"]["modifier"]["list_long"] = []
    spec_dict["functions"]["modifier"]["list_short"] = []

    spec_dict["functions"]["to_short"] = {}
    spec_dict["functions"]["to_long"] = {}

    for func_name in spec_dict["functions"]["info"]:

        abbreviated_name = spec_dict["functions"]["info"][func_name]["abbreviation"]

        spec_dict["functions"]["list"].extend((func_name, abbreviated_name))

        spec_dict["functions"]["list_long"].append(func_name)
        spec_dict["functions"]["list_short"].append(abbreviated_name)

        if spec_dict["functions"]["info"][func_name]["type"] == "primary":
            spec_dict["functions"]["primary"]["list_long"].append(func_name)
            spec_dict["functions"]["primary"]["list_short"].append(abbreviated_name)
        else:
            spec_dict["functions"]["modifier"]["list_long"].append(func_name)
            spec_dict["functions"]["modifier"]["list_short"].append(abbreviated_name)

        spec_dict["functions"]["to_short"][abbreviated_name] = abbreviated_name
        spec_dict["functions"]["to_short"][func_name] = abbreviated_name

        spec_dict["functions"]["to_long"][abbreviated_name] = func_name
        spec_dict["functions"]["to_long"][func_name] = func_name

    return spec_dict


def add_namespaces(spec_dict):
    """Add namespace convenience keys, list, list_{short|long}, to_{short|long}"""

    for ns in spec_dict["namespaces"]:
        spec_dict["namespaces"][ns]["list"] = []
        spec_dict["namespaces"][ns]["list_long"] = []
        spec_dict["namespaces"][ns]["list_short"] = []

        spec_dict["namespaces"][ns]["to_short"] = {}
        spec_dict["namespaces"][ns]["to_long"] = {}

        for obj in spec_dict["namespaces"][ns]["info"]:
            spec_dict["namespaces"][ns]["list"].extend([obj["name"], obj["abbreviation"]])
            spec_dict["namespaces"][ns]["list_short"].append(obj["abbreviation"])
            spec_dict["namespaces"][ns]["list_long"].append(obj["name"])

            spec_dict["namespaces"][ns]["to_short"][obj["abbreviation"]] = obj["abbreviation"]
            spec_dict["namespaces"][ns]["to_short"][obj["name"]] = obj["abbreviation"]

            spec_dict["namespaces"][ns]["to_long"][obj["abbreviation"]] = obj["name"]
            spec_dict["namespaces"][ns]["to_long"][obj["name"]] = obj["name"]

            # For AminoAcid namespace
            if "abbrev1" in obj:
                spec_dict["namespaces"][ns]["to_short"][obj["abbrev1"]] = obj["abbreviation"]
                spec_dict["namespaces"][ns]["to_long"][obj["abbrev1"]] = obj["name"]


def enhance_function_signatures(spec_dict: Mapping[str, Any]) -> Mapping[str, Any]:
    """Enhance function signatures

    Add required and optional objects to signatures objects for semantic validation
    support.

    Args:
        spec_dict (Mapping[str, Any]): bel specification dictionary

    Returns:
        Mapping[str, Any]: return enhanced bel specification dict
    """

    for func in spec_dict["functions"]["signatures"]:
        for i, sig in enumerate(spec_dict["functions"]["signatures"][func]["signatures"]):
            args = sig["arguments"]
            req_args = []
            pos_args = []
            opt_args = []
            mult_args = []
            for arg in args:
                # Multiple argument types
                if arg.get("multiple", False):
                    if arg["type"] in ["Function", "Modifier"]:
                        mult_args.extend(arg.get("values", []))
                    elif arg["type"] in ["StrArgNSArg", "NSArg", "StrArg"]:
                        # Complex signature has this
                        mult_args.append(arg["type"])

                # Optional, position dependent - will be added after req_args based on order in bel_specification
                elif arg.get("optional", False) and arg.get("position", False):
                    if arg["type"] in ["Function", "Modifier"]:
                        pos_args.append(arg.get("values", []))
                    elif arg["type"] in ["StrArgNSArg", "NSArg", "StrArg"]:
                        pos_args.append(arg["type"])

                # Optional, position independent
                elif arg.get("optional", False):
                    if arg["type"] in ["Function", "Modifier"]:
                        opt_args.extend(arg.get("values", []))
                    elif arg["type"] in ["StrArgNSArg", "NSArg", "StrArg"]:
                        opt_args.append(arg["type"])

                # Required arguments, position dependent
                else:
                    if arg["type"] in ["Function", "Modifier"]:
                        req_args.append(arg.get("values", []))
                    elif arg["type"] in ["StrArgNSArg", "NSArg", "StrArg"]:
                        req_args.append(arg["type"])

            spec_dict["functions"]["signatures"][func]["signatures"][i]["req_args"] = copy.deepcopy(
                req_args
            )
            spec_dict["functions"]["signatures"][func]["signatures"][i]["pos_args"] = copy.deepcopy(
                pos_args
            )
            spec_dict["functions"]["signatures"][func]["signatures"][i]["opt_args"] = copy.deepcopy(
                opt_args
            )
            spec_dict["functions"]["signatures"][func]["signatures"][i][
                "mult_args"
            ] = copy.deepcopy(mult_args)

    return spec_dict


def get_ebnf_template():
    """Get EBNF template from Github belbio/bel_specifications repo"""

    spec_dir = config["bel"]["lang"]["specifications"]
    local_fp = f"{spec_dir}/bel.ebnf.j2"

    repo_url = (
        "https://api.github.com/repos/belbio/bel_specifications/contents/resources/bel.ebnf.j2"
    )

    params = {}
    github_access_token = os.getenv("GITHUB_ACCESS_TOKEN", "")
    if github_access_token:
        params = {"access_token": github_access_token}

    try:
        # Get download url for template file
        r = requests.get(repo_url, params=params)

        if r.status_code == 200:
            template_url = r.json()["download_url"]
        else:
            log.warning("Could not get EBNF file download url from Github")

        # Get template file
        try:
            r = requests.get(template_url, params=params, allow_redirects=True)
            if r.status_code == 200:
                open(local_fp, "wt").write(r.text)
            else:
                log.warning(
                    f"Could not download EBNF file from Github -- Status: {r.status_code}  Msg: {r.text}"
                )

        except Exception as e:
            log.warning(
                f"Could not download EBNF file from Github -- Status: {r.status_code}  Msg: {e}"
            )

    except Exception as e:
        log.warning("Could not download BEL EBNF template file")
        if not os.path.exists(f"{spec_dir}/local_fp"):
            log.error("No BEL EBNF template file available")

    return local_fp


def create_ebnf_parser(files):
    """Create EBNF files and EBNF-based parsers"""

    flag = False
    for belspec_fn in files:
        # Get EBNF Jinja template from Github if enabled
        if config["bel"]["lang"]["specification_github_repo"]:
            tmpl_fn = get_ebnf_template()

        # Check if EBNF file is more recent than belspec_fn
        ebnf_fn = belspec_fn.replace(".yaml", ".ebnf")
        if not os.path.exists(ebnf_fn) or os.path.getmtime(belspec_fn) > os.path.getmtime(ebnf_fn):
            with open(belspec_fn, "r") as f:
                belspec = yaml.load(f, Loader=yaml.SafeLoader)

            tmpl_dir = os.path.dirname(tmpl_fn)
            tmpl_basename = os.path.basename(tmpl_fn)

            bel_major_version = belspec["version"].split(".")[0]

            env = jinja2.Environment(
                loader=jinja2.FileSystemLoader(tmpl_dir)
            )  # create environment for template
            template = env.get_template(tmpl_basename)  # get the template

            # replace template placeholders with appropriate variables
            relations_list = [
                (relation, belspec["relations"]["info"][relation]["abbreviation"])
                for relation in belspec["relations"]["info"]
            ]
            relations_list = sorted(list(itertools.chain(*relations_list)), key=len, reverse=True)

            functions_list = [
                (function, belspec["functions"]["info"][function]["abbreviation"])
                for function in belspec["functions"]["info"]
                if belspec["functions"]["info"][function]["type"] == "primary"
            ]
            functions_list = sorted(list(itertools.chain(*functions_list)), key=len, reverse=True)

            modifiers_list = [
                (function, belspec["functions"]["info"][function]["abbreviation"])
                for function in belspec["functions"]["info"]
                if belspec["functions"]["info"][function]["type"] == "modifier"
            ]
            modifiers_list = sorted(list(itertools.chain(*modifiers_list)), key=len, reverse=True)

            created_time = datetime.datetime.now().strftime("%B %d, %Y - %I:%M:%S%p")

            ebnf = template.render(
                functions=functions_list,
                m_functions=modifiers_list,
                relations=relations_list,
                bel_version=belspec["version"],
                bel_major_version=bel_major_version,
                created_time=created_time,
            )

            with open(ebnf_fn, "w") as f:
                f.write(ebnf)

            parser_fn = ebnf_fn.replace(".ebnf", "_parser.py")

            parser = tatsu.to_python_sourcecode(ebnf, filename=parser_fn)
            flag = True
            with open(parser_fn, "wt") as f:
                f.write(parser)

    if flag:
        # In case we created new parser modules
        importlib.invalidate_caches()


def get_function_help(function: str, bel_spec: BELSpec):
    """Get function_help given function name

    This will get the function summary template (argument summary in signature)
    and the argument help listing.
    """

    function_long = bel_spec["functions"]["to_long"].get(function)
    function_help = []

    if function_long:
        for signature in bel_spec["functions"]["signatures"][function_long]["signatures"]:
            function_help.append(
                {
                    "function_summary": signature["argument_summary"],
                    "argument_help": signature["argument_help_listing"],
                    "description": bel_spec["functions"]["info"][function_long]["description"],
                }
            )

    return function_help


def main():
    pass
    # import timy
    # with timy.Timer() as timer:
    #     belspec_yaml2json()
    #     print(timer)


if __name__ == "__main__":
    main()

else:
    # If building documents in readthedocs - provide empty config
    if os.getenv("READTHEDOCS", False):
        config = {}
        log.info("READTHEDOCS environment")
    else:
        update_specifications()



# Compute Edges related code

import copy
from typing import List, Any, Mapping, MutableSequence

from bel.lang.ast import BELAst, Function, NSArg

from structlog import get_logger

log = get_logger()

# Typings
Edges = MutableSequence[Mapping[str, Any]]
BELSpec = Mapping[str, Any]
ComputeRules = List[str]

locations = {
    "extracellular": NSArg("GO", "extracellular space"),
    "cellsurface": NSArg("GO", "cell surface"),
}


def compute_edges(ast: BELAst, spec: BELSpec) -> Edges:
    """Compute edges"""

    edges = []
    if ast.bel_object.__class__.__name__ == "BELAst":
        edges.append(ast.bel_object)

    process_ast(edges, ast, spec)
    return edges


def process_ast(edges, ast, spec):

    if isinstance(ast, BELAst):
        pass

    # TODO composite is not being addressed right now
    # TODO
    elif isinstance(ast, Function):
        if ast.name in ("complex", "complexAbundance"):
            for arg in ast.args:
                edges.append(BELAst(ast, "hasComponent", arg, spec))

        elif ast.name in ("composite", "compositeAbundance"):
            for arg in ast.args:
                edges.append(BELAst(ast, "hasAssociation", arg, spec))

        elif ast.name in ("act", "activity"):
            subject = ast.args[0]
            edge = BELAst(subject, "hasActivity", ast, spec)
            edges.append(edge)

        elif ast.name in ("pmod", "proteinModification"):
            parent = ast.parent_function
            src_abundance = Function(parent.name, spec)
            src_abundance.add_argument(parent.args[0])
            edges.append(BELAst(src_abundance, "hasModification", parent, spec))

        elif ast.name in ("var", "variant"):
            parent = ast.parent_function
            src_abundance = Function(parent.name, spec)
            src_abundance.add_argument(parent.args[0])
            edges.append(BELAst(src_abundance, "hasVariant", parent, spec))

        elif ast.name in ("frag", "fragment"):
            parent = ast.parent_function
            src_abundance = Function(parent.name, spec)
            src_abundance.add_argument(parent.args[0])
            edges.append(BELAst(src_abundance, "hasFragment", parent, spec))

        elif ast.name in ("loc", "location"):
            parent = ast.parent_function
            src_abundance = Function(parent.name, spec)
            src_abundance.add_argument(parent.args[0])
            edges.append(BELAst(src_abundance, "hasLocation", parent, spec))

        elif ast.name in ("tloc", "translocation"):
            src_abundance = ast.args[0]
            from_abundance = copy.deepcopy(src_abundance)
            from_loc = Function("loc", spec)
            from_loc.add_argument(ast.args[1].args[0])
            from_abundance.add_argument(from_loc)

            to_abundance = copy.deepcopy(src_abundance)
            to_loc = Function("loc", spec)
            to_loc.add_argument(ast.args[2].args[0])
            to_abundance.add_argument(to_loc)

            edges.append(BELAst(ast, "decreases", from_abundance, spec))
            edges.append(BELAst(ast, "increases", to_abundance, spec))
            edges.append(BELAst(src_abundance, "hasLocation", from_abundance, spec))
            edges.append(BELAst(src_abundance, "hasLocation", to_abundance, spec))

        elif ast.name in ("sec", "cellSecretion", "surf", "cellSurfaceExpression"):
            target_loc = locations["extracellular"]
            if ast.name in ("surf", "cellSurfaceExpression"):
                target_loc = locations["cellsurface"]

            src_abundance = ast.args[0]
            to_abundance = copy.deepcopy(src_abundance)
            to_loc = Function("loc", spec)
            to_loc.add_argument(target_loc)
            to_abundance.add_argument(to_loc)
            edges.append(BELAst(ast, "increases", to_abundance, spec))
            edges.append(BELAst(src_abundance, "hasLocation", to_abundance, spec))

        elif ast.name in ("deg", "degradation"):
            edges.append(BELAst(ast, "directlyDecreases", ast.args[0], spec))

        elif ast.name in ("fus", "fusion"):
            parent = ast.parent_function

            src_abundance = Function(parent.name, spec)
            src_abundance.add_argument(ast.args[0])
            edges.append(BELAst(src_abundance, "hasFusion", parent, spec))

            src_abundance = Function(parent.name, spec)
            src_abundance.add_argument(ast.args[2])
            edges.append(BELAst(src_abundance, "hasFusion", parent, spec))

        elif ast.name in ("reactants", "products"):
            parent = ast.parent_function

            relation = "hasProduct"
            if ast.name == "reactants":
                relation = "hasReactant"

            for arg in ast.args:
                edges.append(BELAst(parent, relation, arg, spec))

        elif ast.name in ("product", "fusion"):
            parent = ast.parent_function

            src_abundance = Function(parent.name, spec)
            src_abundance.add_argument(ast.args[0])
            edges.append(BELAst(src_abundance, "hasFusion", parent, spec))

            src_abundance = Function(parent.name, spec)
            src_abundance.add_argument(ast.args[2])
            edges.append(BELAst(src_abundance, "hasFusion", parent, spec))

    # Recursively process every element by processing BELAst and Functions
    if hasattr(ast, "args"):
        for arg in ast.args:
            process_ast(edges, arg, spec)


# TODO - not used anymore - review to see if we want to take notes on the approach
def process_rule(edges: Edges, ast: Function, rule: Mapping[str, Any], spec: BELSpec):
    """Process computed edge rule

    Recursively processes BELAst versus a single computed edge rule

    Args:
        edges (List[Tuple[Union[Function, str], str, Function]]): BEL Edge ASTs
        ast (Function): BEL Function AST
        rule (Mapping[str, Any]: computed edge rule
    """
    ast_type = ast.__class__.__name__
    trigger_functions = rule.get("trigger_function", [])
    trigger_types = rule.get("trigger_type", [])
    rule_subject = rule.get("subject")
    rule_relation = rule.get("relation")
    rule_object = rule.get("object")

    log.debug(f"Running {rule_relation}  Type: {ast_type}")

    if isinstance(ast, Function):
        function_name = ast.name
        args = ast.args
        parent_function = ast.parent_function

        if function_name in trigger_functions:
            if rule_subject == "trigger_value":
                subject = ast

            if rule_object == "args":
                for arg in args:
                    log.debug(f"1: {subject} {arg}")
                    edge_ast = BELAst(subject, rule_relation, arg, spec)
                    edges.append(edge_ast)
            elif rule_object == "parent_function" and parent_function:
                log.debug(f"2: {subject} {parent_function}")
                edge_ast = BELAst(subject, rule_relation, parent_function, spec)
                edges.append(edge_ast)

        elif ast_type in trigger_types:
            if rule_subject == "trigger_value":
                subject = ast

            if rule_object == "args":
                for arg in args:
                    log.debug(f"3: {subject} {arg}")
                    edge_ast = BELAst(subject, rule_relation, arg, spec)
                    edges.append(edge_ast)
            elif rule_object == "parent_function" and parent_function:
                log.debug(f"4: {subject} {parent_function}")
                edge_ast = BELAst(subject, rule_relation, parent_function, spec)
                edges.append(edge_ast)

    if isinstance(ast, NSArg):
        term = "{}:{}".format(ast.namespace, ast.value)
        parent_function = ast.parent_function

        if ast_type in trigger_types:
            if rule_subject == "trigger_value":
                subject = term

            if rule_object == "args":
                for arg in args:
                    log.debug(f"5: {subject} {arg}")
                    edge_ast = BELAst(subject, rule_relation, arg, spec)
                    edges.append(edge_ast)
            elif rule_object == "parent_function" and parent_function:
                log.debug(f"6: {subject} {parent_function}")
                edge_ast = BELAst(subject, rule_relation, parent_function, spec)
                edges.append(edge_ast)

    # Recursively process every element by processing BELAst and Functions
    if hasattr(ast, "args"):
        for arg in ast.args:
            process_rule(edges, arg, rule, spec)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This file contains tools used in our parser for BEL statements.
"""

import collections
import math
import os
import pprint
import random
import re
import string
import requests
from typing import Mapping, List, Any

import yaml
import json
import bel.lang.exceptions as bel_ex
from bel.lang.ast import BELAst, Function, Arg, NSArg, StrArg


###################
# STATEMENT TOOLS #
###################


class InvalidStatementObject(object):
    def __init__(self, s, r, o):
        self.subject = s
        self.relation = r
        self.object = o
        self.string_form = self.to_string_form(s, r, o)

    def to_string_form(self, s, r, o):

        sub = self.stmt_creation_decode(s)
        rlt = r
        obj = self.stmt_creation_decode(o)

        return "{} {} {}".format(sub, rlt, obj)

    def stmt_creation_decode(self, ast_dict):

        arg_type = ast_dict.get("type", False)
        arg_value = ast_dict.get("value", False)

        if arg_type == "Function":
            f_name = ast_dict.get("name", None)

            tmp = self.stmt_creation_decode(arg_value[0])

            for arg in arg_value[1:]:
                tmp += ", {}".format(self.stmt_creation_decode(arg))

            return "{}({})".format(f_name, tmp)

        elif arg_type == "String":
            return arg_value
        elif arg_type == "Entity":
            return arg_value
        else:
            return "UNK"


# TODO: move these functions to appropriate place; create_invalid , make_statement, etc. need not be in tools.
def create_invalid(bel_obj, count, max_args):

    list_of_statement_objs = []

    for _ in range(count):
        stmt_obj = make_statement(bel_obj, max_args)
        list_of_statement_objs.append(stmt_obj)

    return list_of_statement_objs


def make_statement(bel_obj, max_args):

    s = choose_and_make_function(bel_obj, max_args)
    r = choose_rand_relation(bel_obj.relations)
    o = choose_and_make_function(bel_obj, max_args)

    return InvalidStatementObject(s, r, o)


def choose_and_make_function(bel_obj, max_args):

    full_func = {
        "name": choose_rand_function(bel_obj.function_signatures),
        "type": "Function",
        "value": [],
    }

    num_args = random.randint(1, max_args)  # how many args to have
    for _ in range(num_args):

        t = random.choice(["Entity", "String", "Function"])
        arg = {"type": t}

        if t == "Entity":
            arg["value"] = random_namespace_arg()
        elif t == "String":
            arg["value"] = random_quoted_string()
        elif t == "Function":
            arg = choose_and_make_function(bel_obj, max_args)
        else:
            pass

        full_func["value"].append(arg)

    return full_func


def choose_rand_relation(relations):

    return random.choice(relations)


def choose_rand_function(func_sigs):

    return random.choice(list(func_sigs.keys()))


def random_namespace_arg():
    ascii_letters = string.ascii_uppercase + string.ascii_lowercase
    ascii_alphanumeric_upper = string.ascii_uppercase + string.digits
    ascii_alphanumeric = ascii_letters + string.digits

    i = random.randint(2, 5)
    rand_nspace = "NS" + "".join(
        random.choice(ascii_alphanumeric_upper) for _ in range(i)
    )

    j = random.randint(5, 25)

    if random.random() < 0.5:  # quoted nsvalue
        rand_nsvalue = "".join(
            random.choice(ascii_alphanumeric + " ' - , + /.") for _ in range(j)
        )
        rand_nsvalue = '"{}"'.format(rand_nsvalue)
    else:  # unquoted nsvalue
        rand_nsvalue = "".join(random.choice(ascii_alphanumeric) for _ in range(j))

    return "{}: {}".format(rand_nspace, rand_nsvalue)


def random_quoted_string():
    ascii_letters = string.ascii_uppercase + string.ascii_lowercase
    ascii_alphanumeric = ascii_letters + string.digits

    j = random.randint(5, 25)

    rand_nsvalue = "".join(
        random.choice(ascii_alphanumeric + " ' - , + /.") for _ in range(j)
    )
    return '"{}"'.format(rand_nsvalue)


# def simple_args(args, bel_obj):
#
#     new_args = []
#
#     for p in args:
#
#         if 'function' in p:
#             obj_type = 'function'
#             f_type = 'primary'
#             obj = {
#                 'name': p['function'],
#                 'type': f_type,
#                 'alternate': bel_obj.translate_terms[p['function']],
#                 'full_string': decode(p),
#                 'args': simple_args(p['function_args'], bel_obj)
#             }
#         elif 'm_function' in p:
#             obj_type = 'function'
#             f_type = 'modifier'
#             obj = {
#                 'name': p['m_function'],
#                 'type': f_type,
#                 'alternate': bel_obj.translate_terms[p['m_function']],
#                 'full_string': decode(p),
#                 'args': simple_args(p['m_function_args'], bel_obj)
#             }
#         elif 'ns_arg' in p:
#             obj_type = 'ns_variable'
#             obj = {
#                 'namespace': p['ns_arg']['nspace'],
#                 'value': p['ns_arg']['ns_value'],
#                 'full_string': decode(p)
#             }
#         else:
#             obj_type = 'str_variable'
#             obj = {
#                 'value': p['str_arg'],
#             }
#
#         new_args.append({obj_type: obj})
#
#     return new_args


# def extract_args_from_rule(rule, function_obj):
#
#     args = []
#
#     print('OLD RULE: {}'.format(rule))
#     rule = rule.replace('{{ ', '').replace(' }}', '')
#
#     # instead of replacing the keyword with the given variable, we need to actually create an object....
#
#     rule = rule.replace('p_full', variables['p_full'])
#     rule = rule.replace('p_name', variables['p_name'])
#
#     rule = rule.replace('full', variables['full'])  # this should be the func object string
#     rule = rule.replace('fn_name', variables['fn_name'])  # this should be the function object
#     print('NEW RULE: {}'.format(rule))
#     print()
#
#     args_wanted = []
#
#     arg_pattern = '(p_)?args(\[[fmns]\])?'
#     regex_pattern = re.compile(arg_pattern)
#     arg_matched_rule = regex_pattern.search(rule)
#
#     if arg_matched_rule:  # if args are needed, loop through
#         final_to_replace = arg_matched_rule.group()
#         use_parent_args = None
#
#         try:
#             use_parent_args = arg_matched_rule.group(1)
#             filter_string = arg_matched_rule.group(2)
#             if filter_string is None:
#                 filter_string = ''
#         except IndexError as e:
#             filter_string = ''  # stands for all
#
#         # print('Use parent args: {}'.format(bool(use_parent_args)))
#         # print('Filter string: {}'.format(filter_string))
#
#         allowed_type = ''
#
#         if 'f' in filter_string:
#             allowed_type = 'function'
#         elif 'm' in filter_string:
#             allowed_type = 'm_function'
#         elif 'n' in filter_string:
#             allowed_type = 'ns_arg'
#         elif 's' in filter_string:
#             allowed_type = 'str_arg'
#
#         if use_parent_args is not None:  # use the parent's args
#             args_to_loop = variables['p_args']
#         else:
#             args_to_loop = args
#
#         for a in args_to_loop:
#             if allowed_type == '' or allowed_type in list(a):
#                 decoded_arg = decode(a)
#                 final_rule = rule.replace(final_to_replace, decoded_arg)
#                 args_wanted.append(final_rule)
#                 make_simple_ast(a)
#
#     else:
#         return [rule]
#
#     return args_wanted


#################
# PARSING TOOLS #
#################


class Colors:
    PINK = "\033[95m"
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"
    END = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
ast_dict from Tatsu is converted to this Ast Class object and used for
validation and BEL transformation (e.g. canonicalization, orthologization, etc)
"""

from typing import Mapping, Any
import traceback
import sys

import bel.lang.bel_utils

import structlog

log = structlog.getLogger(__name__)


########################
# BEL statement AST #
########################
class BELAst(object):
    def __init__(self, bel_subject, bel_relation, bel_object, spec):
        self.bel_subject = bel_subject
        self.bel_relation = bel_relation
        self.bel_object = bel_object
        self.spec = spec  # bel specification dictionary
        self.type = "BELAst"
        self.species = set()  # tuples of (species_id, species_label)
        self.collected_nsarg_norms = False
        self.collected_orthologs = False
        self.partially_orthologized = False
        self.args = [self.bel_subject, self.bel_object]

    def dump(self):
        return {
            "bel_subject": self.bel_subject.to_string(),
            "bel_relation": self.bel_relation,
            "bel_object": self.bel_object.to_string(),
            "type": self.type,
            "collected_nsarg_norms": self.collected_nsarg_norms,
            "collected_orthologs": self.collected_orthologs,
            "args": self.args,
            "spec": "Not included",
        }

    def canonicalize(self):

        if not self.collected_nsarg_norms:
            log.error(
                f"Cannot canonicalize without running collected_nsarg_norms() on BEL object first {self.dump()}"
            )
            log.info(self.dump)
            traceback.print_stack(file=sys.stdout)
            return self

        if self and isinstance(self, NSArg):
            self.canonicalize()

        if hasattr(self, "args"):
            for arg in self.args:
                if arg:
                    arg.canonicalize()

        return self

    def decanonicalize(self):
        if not self.collected_nsarg_norms:
            log.error(
                f"Cannot decanonicalize without running collected_nsarg_norms() on BEL object first {self.dump()}"
            )
            log.info(self.dump)
            traceback.print_stack(file=sys.stdout)
            return self

        if self and isinstance(self, NSArg):
            self.decanonicalize()

        if hasattr(self, "args"):
            for arg in self.args:
                if arg:
                    arg.decanonicalize()

        return self

    def orthologize(self, species_id, belast=None):
        if not self.collected_orthologs:
            log.error(
                f"Cannot orthologize without running collect_orthologs() on BEL object first {self.dump()}"
            )
            log.info(self.dump)
            traceback.print_stack(file=sys.stdout)
            return self

        if hasattr(self, "args"):
            if belast is None:
                belast = self
            for arg in self.args:
                if arg:
                    arg.orthologize(species_id, belast)

        return self

    def to_string(self, ast_obj=None, fmt: str = "medium") -> str:
        """Convert AST object to string

        Args:
            fmt (str): short, medium, long formatted BEL statements
                short = short function and short relation format
                medium = short function and long relation format
                long = long function and long relation format
            canonicalize

        Returns:
            str: string version of BEL AST
        """

        if not ast_obj:
            ast_obj = self

        bel_relation = None
        if self.bel_relation and fmt == "short":
            bel_relation = self.spec["relations"]["to_short"].get(
                self.bel_relation, self.bel_relation
            )
        elif self.bel_relation:
            bel_relation = self.spec["relations"]["to_long"].get(
                self.bel_relation, self.bel_relation
            )

        if self.bel_subject and bel_relation and self.bel_object:
            if isinstance(self.bel_object, BELAst):
                return "{} {} ({})".format(
                    self.bel_subject.to_string(fmt=fmt),
                    bel_relation,
                    self.bel_object.to_string(fmt=fmt),
                )
            else:
                return "{} {} {}".format(
                    self.bel_subject.to_string(fmt=fmt),
                    bel_relation,
                    self.bel_object.to_string(fmt=fmt),
                )

        elif self.bel_subject:
            return "{}".format(self.bel_subject.to_string(fmt=fmt))

        else:
            return ""

    def to_triple(self, ast_obj=None, fmt="medium"):
        """Convert AST object to BEL triple

        Args:
            fmt (str): short, medium, long formatted BEL statements
                short = short function and short relation format
                medium = short function and long relation format
                long = long function and long relation format

        Returns:
            dict: {'subject': <subject>, 'relation': <relations>, 'object': <object>}
        """

        if not ast_obj:
            ast_obj = self

        if self.bel_subject and self.bel_relation and self.bel_object:
            if self.bel_relation.startswith("has"):
                bel_relation = self.bel_relation
            elif fmt == "short":
                bel_relation = self.spec["relations"]["to_short"].get(
                    self.bel_relation, None
                )
            else:
                bel_relation = self.spec["relations"]["to_long"].get(
                    self.bel_relation, None
                )

            bel_subject = self.bel_subject.to_string(fmt=fmt)

            if isinstance(self.bel_object, (BELAst)):
                bel_object = f"({self.bel_object.to_string(fmt=fmt)})"
            else:
                bel_object = self.bel_object.to_string(fmt=fmt)

            return {
                "subject": bel_subject,
                "relation": bel_relation,
                "object": bel_object,
            }

        elif self.bel_subject:
            return {"subject": self.bel_subject.to_string(fmt=fmt)}

        else:
            return None

    def __str__(self):
        return self.to_string(self)

    __repr__ = __str__

    def print_tree(self, ast_obj=None):
        """Convert AST object to tree view of BEL AST

        Returns:
            prints tree of BEL AST to STDOUT
        """

        if not ast_obj:
            ast_obj = self

        if hasattr(self, "bel_subject"):
            print("Subject:")
            self.bel_subject.print_tree(self.bel_subject, indent=0)

        if hasattr(self, "bel_relation"):
            print("Relation:", self.bel_relation)

        if hasattr(self, "bel_object"):
            if self.bel_object.type == "BELAst":
                if hasattr(self, "bel_subject"):
                    print("Nested Subject:")
                    self.bel_object.bel_subject.print_tree(indent=0)

                if hasattr(self, "bel_relation"):
                    print("Nested Relation:", self.bel_object.bel_relation)

                if hasattr(self, "bel_object"):
                    print("Nested Object:")
                    self.bel_object.bel_object.print_tree(indent=0)
            else:
                print("Object:")
                self.bel_object.print_tree(self.bel_object, indent=0)

        return self


###################
# Function object #
###################


class Function(object):
    def __init__(self, name, spec, parent_function=None):

        self.name = spec["functions"]["to_long"].get(name, name)
        self.name_short = spec["functions"]["to_short"].get(name, name)
        if self.name in spec["functions"]["info"]:
            self.function_type = spec["functions"]["info"][self.name]["type"]
        else:
            self.function_type = ""

        self.type = "Function"

        self.parent_function = parent_function
        self.spec = spec
        self.position_dependent = False
        self.args = []
        self.siblings = []

    def is_primary(self):
        if self.function_type == "primary":
            return True
        return False

    def is_modifier(self):
        if self.function_type == "modifier":
            return True
        return False

    def add_argument(self, arg):
        self.args.append(arg)

    def add_sibling(self, sibling):
        self.siblings.append(sibling)

    def change_parent_fn(self, parent_function):
        self.parent_function = parent_function

    def change_function_type(self, function_type):
        self.function_type = function_type

    def canonicalize(self):
        if isinstance(self, NSArg):
            self.canonicalize()

        if hasattr(self, "args"):
            for arg in self.args:
                arg.canonicalize()

        return self

    def decanonicalize(self):
        if isinstance(self, NSArg):
            self.decanonicalize()

        if hasattr(self, "args"):
            for arg in self.args:
                arg.decanonicalize()

        return self

    def orthologize(self, species_id, belast):
        if isinstance(self, NSArg):
            self.orthologize(species_id)

        if hasattr(self, "args"):
            for arg in self.args:
                arg.orthologize(species_id, belast)

        return self

    def to_string(
        self,
        fmt: str = "medium",
        canonicalize: bool = False,
        decanonicalize: bool = False,
        orthologize: str = None,
    ) -> str:
        """Convert AST object to string

        Args:
            fmt (str): short, medium, long formatted BEL statements
                short = short function and short relation format
                medium = short function and long relation format
                long = long function and long relation format

        Returns:
            str: string version of BEL AST
        """

        arg_string = ", ".join([a.to_string(fmt=fmt) for a in self.args])

        if fmt in ["short", "medium"]:
            function_name = self.name_short
        else:
            function_name = self.name

        return "{}({})".format(function_name, arg_string)

    def __str__(self):
        arg_string = ", ".join([a.to_string() for a in self.args])
        return "{}({})".format(self.name, arg_string)

    __repr__ = __str__

    def print_tree(self, ast_obj=None, indent=0):
        if not ast_obj:
            ast_obj = self
        for arg in self.args:
            if arg.__class__.__name__ == "Function":
                arg.print_tree(arg, indent + 1)
            else:
                print("\t" * (indent + 1) + arg.print_tree())

    def subcomponents(self, subcomponents):
        """Generate subcomponents of the BEL subject or object

        These subcomponents are used for matching parts of a BEL
        subject or Object in the Edgestore.

        Args:
            AST
            subcomponents:  Pass an empty list to start a new subcomponents request

        Returns:
            List[str]: subcomponents of BEL subject or object
        """

        for arg in self.args:
            if arg.__class__.__name__ == "Function":
                subcomponents.append(arg.to_string())
                if arg.function_type == "primary":
                    arg.subcomponents(subcomponents)
            else:
                subcomponents.append(arg.to_string())

        return subcomponents


#####################
# Argument objects #
#####################
class Arg(object):
    def __init__(self, parent_function):
        self.parent_function = parent_function
        self.siblings = []
        self.optional = False
        self.type = "Arg"

    def add_sibling(self, sibling):
        self.siblings.append(sibling)

    def canonicalize(self):
        return self

    def decanonicalize(self):
        return self

    def orthologize(self, species_id, belast):
        return self


class NSArg(Arg):
    def __init__(self, namespace, value, parent_function=None, value_types=[]):
        Arg.__init__(self, parent_function)
        self.namespace = namespace
        self.value = self.normalize_nsarg_value(value)
        self.value_types = value_types
        self.type = "NSArg"
        self.canonical = None
        self.decanonical = None
        self.species_id = None
        self.species_label = None
        self.orthologs = (
            {}
        )  # {'TAX:9606': {'species_label': 'human', 'canonical': 'EG:207', 'decanonical': 'HGNC:AKT1'}, 'TAX:10090': {'species_label': 'mouse', canonical': 'EG:11651', 'decanonical': 'MGI:Akt1'}, ...
        self.orthology_species = None
        self.orthologized = (
            None
        )  # True for orthologized - False -> unable to orthologize
        self.original = f"{namespace}:{value}"

        # What entity types can this be from the function signatures?
        #    this is used for statement autocompletion and entity validation
        self.entity_types = []

    def change_nsvalue(self, namespace, value):
        """Deprecated"""

        self.namespace = namespace
        self.value = value

    def update_nsval(
        self, *, nsval: str = None, ns: str = None, val: str = None
    ) -> None:
        """Update Namespace and valueast.

        Args:
            nsval: e.g. HGNC:AKT1
            ns: namespace
            val: value of entity
        """

        if not (ns and val) and nsval:
            (ns, val) = nsval.split(":", 1)
        elif not (ns and val) and not nsval:
            log.error("Did not update NSArg - no ns:val or nsval provided")

        self.namespace = ns
        self.value = val

    def add_value_types(self, value_types):
        self.value_types = value_types

    def normalize_nsarg_value(self, nsarg_value):
        """Normalize NSArg value

        If needs quotes (only if it contains whitespace, comma or ')' ), make sure
        it is quoted, else remove quotes

        Args:
            nsarg_value (str): NSArg value, e.g. AKT1 of HGNC:AKT1

        Returns:
            str:
        """

        return bel.lang.bel_utils.quoting_nsarg(nsarg_value)

    def canonicalize(self):
        if isinstance(self, NSArg):
            self.update_nsval(nsval=self.canonical)
        return self

    def decanonicalize(self):
        if isinstance(self, NSArg):
            self.update_nsval(nsval=self.decanonical)
        return self

    def orthologize(self, ortho_species_id, belast):
        """Decanonical ortholog name used"""

        if (
            self.orthologs
            and ortho_species_id in self.orthologs
            and ortho_species_id != self.species_id
        ):
            self.orthology_species = ortho_species_id
            self.canonical = self.orthologs[ortho_species_id]["canonical"]
            self.decanonical = self.orthologs[ortho_species_id]["decanonical"]
            self.update_nsval(nsval=self.decanonical)
            self.orthologized = True

        elif self.species_id and ortho_species_id not in self.orthologs:
            self.orthologized = False
            belast.partially_orthologized = True

        return self

    def to_string(self, fmt: str = "medium") -> str:
        """Convert AST object to string

        Args:
            fmt (str): short, medium, long formatted BEL statements
                short = short function and short relation format
                medium = short function and long relation format
                long = long function and long relation format

        Returns:
            str: string version of BEL AST
        """
        return f"{self.namespace}:{self.value}"

    def print_tree(self, fmt: str = "medium") -> str:
        return f"NSArg: {self.namespace}:{self.value} canonical: {self.canonical} decanonical: {self.decanonical} orthologs: {self.orthologs} orig_species: {self.orthology_species}"

    def __str__(self):
        return f"{self.namespace}:{self.value}"

    __repr__ = __str__


class StrArg(Arg):
    def __init__(self, value, parent_function, value_types=[]):
        Arg.__init__(self, parent_function)
        self.value = value
        self.value_types = value_types
        self.type = "StrArg"

    def add_value_types(self, value_types):
        self.value_types = value_types

    def to_string(self, fmt: str = "medium") -> str:
        """Convert AST object to string

        Args:
            fmt (str): short, medium, long formatted BEL statements
                short = short function and short relation format
                medium = short function and long relation format
                long = long function and long relation format

        Returns:
            str: string version of BEL AST
        """
        return "{}".format(self.value)

    def print_tree(self, fmt: str = "medium") -> str:
        return "{}".format(self.value)

    def __str__(self):
        return "StrArg: {}".format(self.value)

    __repr__ = __str__


def ast_dict_to_objects(ast_dict: Mapping[str, Any], bel_obj) -> BELAst:
    """Convert Tatsu AST dictionary to BEL AST object

    Args:
        ast_dict (Mapping[str, Any])

    Returns:
        BELAst: object representing the BEL Statement AST
    """
    ast_subject = ast_dict.get("subject", None)
    ast_object = ast_dict.get("object", None)

    bel_subject = None
    bel_object = None
    bel_relation = ast_dict.get("relation")

    if ast_subject:
        bel_subject = function_ast_to_objects(ast_subject, bel_obj)

    if ast_object:
        bel_object = function_ast_to_objects(ast_object, bel_obj)

    ast_obj = BELAst(bel_subject, bel_relation, bel_object, bel_obj.spec)

    return ast_obj


def function_ast_to_objects(fn_ast_dict, bel_obj):
    # needed and used

    # print(fn_ast_dict)
    spec = bel_obj.spec  # bel specification

    func_name = fn_ast_dict.get("function", None)
    potential_bel_stmt = fn_ast_dict.get("bel_statement", None)

    if (
        func_name is None and potential_bel_stmt is not None
    ):  # this is a nested BEL statement
        tmp_fn_obj = ast_dict_to_objects(potential_bel_stmt, bel_obj)
    else:
        tmp_fn_obj = Function(func_name, spec)
        tmp_fn_obj_args = fn_ast_dict.get("function_args", None)

        # for each argument in tmp_fn_obj_args, add it to our function object using add_args_to_compute_obj()
        add_args_to_compute_obj(bel_obj, tmp_fn_obj, tmp_fn_obj_args)

    return tmp_fn_obj


def add_args_to_compute_obj(our_bel_obj, our_obj, our_obj_args):
    # needed and used
    tmp_all_args_objs = []

    for argument in our_obj_args:

        fn_args = []
        tmp_arg_obj = None

        if "function" in argument:
            fn = argument.get("function", None)
            fn_parent = our_obj
            fn_args = argument.get("function_args", [])

            tmp_arg_obj = Function(fn, our_bel_obj.spec, parent_function=fn_parent)

        elif "modifier" in argument:
            fn = argument.get("modifier", None)
            fn_parent = our_obj
            fn_args = argument.get("modifier_args", [])

            tmp_arg_obj = Function(fn, our_bel_obj.spec, parent_function=fn_parent)

        elif "ns_arg" in argument:
            ns = argument["ns_arg"]["ns"]
            nsv = argument["ns_arg"]["ns_value"]
            ns_parent = our_obj

            tmp_arg_obj = NSArg(ns, nsv, ns_parent)

        elif "str_arg" in argument:
            sv = argument.get("str_arg", None)
            sv_parent = our_obj

            tmp_arg_obj = StrArg(sv, sv_parent)

        else:
            continue

        if fn_args:
            add_args_to_compute_obj(our_bel_obj, tmp_arg_obj, fn_args)

        if tmp_arg_obj is not None:
            our_obj.add_argument(tmp_arg_obj)
            tmp_all_args_objs.append(tmp_arg_obj)

    # nested loop to add siblings to each arg. exclude the arg itself as its own sibling by ignoring if index match.
    for idx, arg_obj in enumerate(tmp_all_args_objs):
        for sibling_idx, sibling_arg_obj in enumerate(tmp_all_args_objs):

            if idx == sibling_idx:  # skip adding self as sibling
                continue
            else:
                arg_obj.add_sibling(sibling_arg_obj)

    return
